#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""A class to format the log messages + the logger instance."""

from logging import Formatter, StreamHandler, getLogger, DEBUG


CNRM = '\x1B[0m'
CBLK = '\x1B[30m'
CRED = '\x1B[31m'
CGRN = '\x1B[32m'
CYEL = '\x1B[33m'
CBLU = '\x1B[34m'
CMAG = '\x1B[35m'
CCYN = '\x1B[36m'
CWHT = '\x1B[37m'
BNRM = '\x1B[1m\x1b[00m'
BBLK = '\x1B[1m\x1b[30m'
BRED = '\x1B[1m\x1b[31m'
BGRN = '\x1B[1m\x1b[32m'
BYEL = '\x1B[1m\x1b[33m'
BBLU = '\x1B[1m\x1b[34m'
BMAG = '\x1B[1m\x1b[35m'
BCYN = '\x1B[1m\x1b[36m'
BWHT = '\x1B[1m\x1b[37m'
BOLD = '\x1B[1m'


class ColorFormatter(Formatter):
    """Add color to log messages by overring logging.Formater methods."""

    FORMAT = ("[$BOLD%(filename)-8s: "
              "%(funcName)20s():%(lineno)4s$RESET][%(levelname)-18s] "
              "%(message)s ")
    BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)

    COLOR_SEQ = "\x1B[%dm"

    COLORS = {
        'WARNING': YELLOW,
        'INFO': WHITE,
        'DEBUG': BLUE,
        'CRITICAL': YELLOW,
        'ERROR': RED
    }

    def __init__(self, use_color=True):
        msg = self.formatter_msg(self.FORMAT, use_color)
        Formatter.__init__(self, msg)
        self.use_color = use_color

    def formatter_msg(self, msg, use_color=True):
        if use_color:
            msg = msg.replace("$RESET", CNRM)
            msg = msg.replace("$BOLD", BOLD)
        else:
            msg = msg.replace("$RESET", "")
            msg = msg.replace("$BOLD", "")
        return msg

    def format(self, record):
        levelname = record.levelname
        if self.use_color and levelname in self.COLORS:
            fore_color = 30 + self.COLORS[levelname]
            levelname_color = self.COLOR_SEQ % fore_color + levelname + CNRM
            record.levelname = levelname_color
        return Formatter.format(self, record)


CF = ColorFormatter()
handler = StreamHandler()
handler.setFormatter(CF)
#: The logger which can be used anywhere in the program
lg = getLogger('root')
lg.setLevel(DEBUG)
lg.addHandler(handler)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Classes to connect to databases."""

from sqlite3 import connect, OperationalError
from re import compile
from tipy.lg import lg


class OpenFileError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr('Cannot open file "%s".' % self.value)


class DatabaseConnector(object):
    """Implement methods for accessing the databases.

    G{classtree DatabaseConnector}
    """

    singleQuoteRegex = compile("'")

    def __init__(self, dbname, maxN=1):
        """DababaseConnector creator.

        @param dbname:
            Path to the database file.
        @type dbname: str
        @param maxN:
            The n in the longer database n-grams table.
        @type maxN: int
        """
        self.maxN = maxN
        self.dbname = dbname

    def crt_ngram_table(self, n=1):
        """Creates a table in the database to store n-gram of a given n.

        @param n:
            The n in n-gram. A table called [n]_gram (where [n] is the n
            parameter) will be created.
        @type n: int
        """
        query = 'CREATE TABLE IF NOT EXISTS _{0}_gram ('.format(n)
        unique = ''
        for i in reversed(range(n)):
            if i != 0:
                unique += 'word_{0}, '.format(i)
                query += 'word_{0} TEXT, '.format(i)
            else:
                unique += 'word'
                query += 'word TEXT, count INTEGER, UNIQUE({0}) );'.format(
                    unique)
        self.execute_sql(query)

    def drop_ngram_table(self, n=1):
        """Drop a n-gram table in the database.

        @param n:
            The n in n-gram.
        @type n: int
        """
        query = 'DROP TABLE IF EXISTS _{0}_gram;'.format(n)
        self.execute_sql(query)

    def crt_index(self, n):
        """Create the index for the table storing n-gram of given n.

        @param n:
            The n in n-gram.
        @type n: int
        """
        for i in reversed(range(n)):
            if i != 0:
                query = 'CREATE INDEX idx_{0}_gram_{1} ON'\
                        '_{0}_gram(word_{1});'.format(n, i)
                self.execute_sql(query)

    def dlt_index(self, n):
        """'Drop the index for the table storing n-gram of given n.

        @param n:
            The n in n-gram.
        @type n: int
        """
        for i in reversed(range(n)):
            if i != 0:
                query = 'DROP INDEX IF EXISTS idx_{0}_gram_{1};'.format(
                    n, i)
                self.execute_sql(query)

    def ngrams(self, withCounts=False):
        """Returns all ngrams that are in every tables of the database.

        @note: The result is convert to tuple before being returned because it
               is used as a dictionary key and only immutable types can be used
               as dictionary key, so list cannot.

        @param withCounts:
            Indicate if the ngrams counts (number of occurences) should be
            returned too.
        @type withCounts: bool

        @return:
            The n-grams of each tables of the database.
        @rtype: tuple
        """
        query = 'SELECT '
        for i in reversed(range(self.maxN)):
            if i != 0:
                query += 'word_{0}, '.format(i)
            elif i == 0:
                query += 'word'
        if withCounts:
            query += ', count'
        query += ' FROM _{0}_gram;'.format(self.maxN)
        result = self.execute_sql(query)
        for row in result:
            yield tuple(row)

    def sum_ngrams_occ(self, n):
        """Compute the occurences sum of every n-grams of given n in database.

        @param n:
            The n in n-gram.
        @type n: int

        @return:
            The sum of the number of occurences of every n-grams in the n-grams
            table of given n in the database.
        @rtype: int
        """
        query = 'SELECT SUM(count) from _' + str(n) + '_gram;'
        return self.extract_first_integer(self.execute_sql(query))

    def ngrams_in_table(self, n):
        """Compute the number of n-grams in the n-grams table of given n.

        @param n:
            The n in n-gram.
        @type n: int

        @return:
            The number of n-grams in the n-grams table of given n in the
            database.
        @rtype: int
        """
        table = '_' + str(n) + '_gram'
        query = 'SELECT Count() from _' + str(n) + '_gram;'
        return self.extract_first_integer(self.execute_sql(query))

    def ngram_count(self, ngram):
        """Retrieve the number of occurences of a given ngram in the database.

        @param ngram:
            The n-gram for which number of occurences must be retrieved.
        @type ngram: list

        @return:
            The number of occurences of the n-gram or 0 if the n-gram is not
            in the databade.
        @rtype: int
        """
        query = 'SELECT count FROM _{0}_gram'.format(len(ngram))
        query += self.make_where_clause(ngram) + ';'
        return self.extract_first_integer(self.execute_sql(query))

    def ngram_table_tp(self, ngram, limit=-1):
        """Retrieve the n-gram records which complete the given n-gram.

        For instance, if ngram is::
            ['on', 'the', 'ta']
        Then the returned records would be somthing like::
            [['on', 'the', 'table'  ], 5]
            [['on', 'the', 'take'   ], 1]
            [['on', 'the', 'taskbar'], 1]

        @note: the query makes sure the n-grams are returned in descending order
               according to their number of occurences. This is important
               because the predictors predict() methods which would call this
               method can limit their suggestion number so the most probable
               suggestions (which are based on most frequent n-grams) must be
               append to the list first.

        @param ngram:
            The n-gram words of the record to retrieve in the database.
        @type ngram: list
        @param limit:
            Maximum number of records to retrieve.
        @type limit: int

        @return:
            Return the n-grams records (n-gram words + number of occurences)
            completing the n-gram in the database or an empty list if no n-grams
            have been found.
        @rtype: list
        """
        query = '{0} FROM _{1}_gram {2} ORDER BY count DESC'.format(
            self.make_select_like_clause(len(ngram)), len(ngram),
            self.make_where_like_clause(ngram))
        if limit < 0:
            query += ';'
        else:
            query += ' LIMIT ' + str(limit) + ';'
        return self.execute_sql(query)

    def insert_ngram(self, ngram, count):
        """Insert an n-gram with its number of occurences into the database.

        @param ngram:
            The n-gram to insert.
        @type ngram: list
        @param count:
            The number of occurences of the n-gram to insert. It is usually 1 as
            the n-grams should be inserted in the database only if they are not
            already in it and they are added when they are read on the input
            buffer.
        @type count: int
        """
        query = 'INSERT INTO _{0}_gram {1};'.format(len(ngram),
                self.make_values_clause(ngram, count))
        self.execute_sql(query)

    def update_ngram(self, ngram, count):
        """Update an n-gram number of occurences in the database.

        @warning: The ngram has to be in the database, otherwise this method
            will fail.

        @param ngram:
            The n-gram to update (!!! should be in the database !!!).
        @type ngram: list
        @param count:
            The number of occurences of the n-gram.
        @type count: int
        """
        query = 'UPDATE _{0}_gram SET count = {1}'.format(len(ngram), count)
        query += self.make_where_clause(ngram) + ';'
        self.execute_sql(query)

    def remove_ngram(self, ngram):
        """Remove a given ngram from the databae.

        @warning: The ngram has to be in the database, otherwise this method
            will fail.

        @param ngram:
            The n-gram to delete (!!! should be in the database !!!).
        @type ngram: list
        """
        query = 'DELETE FROM _{0}_gram'.format(len(ngram))
        query += self.make_where_clause(ngram) + ';'
        self.execute_sql(query)

    def make_values_clause(self, ngram, count):
        ng = [self.singleQuoteRegex.sub("''", n) for n in ngram]
        valuesClause = "VALUES('" + "', '".join(ng) + "', {0})".format(count)
        return valuesClause

    def make_where_clause(self, ngram):
        whereClause = " WHERE"
        for i in range(len(ngram)):
            n = self.singleQuoteRegex.sub("''", ngram[i])
            if i < (len(ngram) - 1):
                whereClause += " word_{0} = '{1}' AND".format(
                    len(ngram) - i - 1, n)
            else:
                whereClause += " word = '{0}'".format(n)
        return whereClause

    def make_select_like_clause(self, n):
        result = "SELECT "
        for i in reversed(range(1, n)):
            result += "word_{0}, ". format(i)
        result += "word, count"
        return result

    def make_where_like_clause(self, ngram):
        whereClause = " WHERE"
        for i in range(len(ngram)):
            if i < (len(ngram) - 1):
                whereClause += " word_{0} = '{1}' AND".format(
                    len(ngram) - i - 1, ngram[i])
            else:
                whereClause += " word LIKE '{0}%'".format(ngram[-1])
        return whereClause

    def extract_first_integer(self, table):
        count = 0
        if table and table[0]:
            count = int(table[0][0])
        if not count > 0:
            count = 0
        return count


class SqliteDatabaseConnector(DatabaseConnector):
    """Database connector for sqlite databases.

    G{classtree SqliteDatabaseConnector}
    """

    def __init__(self, dbname, maxN=1):
        """SqliteDatabaseConnector creator.

        @param dbname:
            Path to the database file.
        @type dbname: str
        @param maxN:
            The n in the longer database n-grams table.
        @type maxN: int
        """
        DatabaseConnector.__init__(self, dbname, maxN)
        self.con = None
        self.open_database()

    def commit(self):
        """Send a commit to the database."""
        self.con.commit()

    def open_database(self):
        """Open the database."""
        try:
            self.con = connect(self.dbname)
        except OperationalError:
            lg.error("Cannot open database file '%s'" % (self.dbname))
            raise OpenFileError(self.dbname)

    def close_database(self):
        """Close the database."""
        if self.con:
            self.con.close()

    def execute_sql(self, query):
        """Execute a query string on an open database.

        @param query:
            The query to execute.
        @type query: str

        @return:
            The query result.
        @rtype: list
        """
        c = self.con.cursor()
        c.execute(query)
        result = c.fetchall()
        return result


def insert_ngrams(ngramMap, n, outfile, append=False, createIndex=False,
                  callback=None):
    """Insert every n-grams of the map in the database.

    This function open a database and insert or update every n-grams of the
    given list in it.

    @param ngramMap:
        The list of n-grams to insert or update.
    @type ngramMap: list
    @param n:
        The n in n-gram.
    @type n: int
    @param outfile:
        Path to the database.
    @type outfile: str
    @param append:
        Indicate weither the n-gram should be append to the database. If the
        database isn't empty, this function could raise conflict when trying
        to insert an n-gram which is already in the database. When such
        conflicts happen the append value is important:
            - True: conflicting records count values are updated.
            - False: conflicting records are replaced by new records.
    @type append: bool
    @param createIndex:
        Indicate weither the database table indexes should be created.
    @type createIndex: bool
    @param callback:
        The callback is used to show the progress percentage. In the gui a
        callback method is implemented to update a progress bar showing the
        n-grams insertion progress (cf. py).
    @type callback: fun(float, ...)
    """
    progress = 0
    sql = SqliteDatabaseConnector(outfile, n)
    sql.crt_ngram_table(n)
    for ngram, count in ngramMap.items():
        if append:
            oldCount = sql.ngram_count(ngram)
            if oldCount > 0:
                sql.update_ngram(ngram, oldCount + count)
            else:
                sql.insert_ngram(ngram, count)
        else:
            sql.remove_ngram(ngram)     # avoid IntegrityError
            sql.insert_ngram(ngram, count)
        progress += 100 / len(ngramMap)
        if callback:
            callback(progress)
    sql.commit()
    if createIndex and not append:
        sql.crt_index(n)
    sql.close_database()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""The Stoplist class holds a list of (undesired) words."""


class StopList(object):
    """Stoplist classes holding every words of the stoplist(s).

    The program's predictors test every suggested words against the stoplist. If
    a suggested word is a stoplist word then the suggested word is not added to
    the suggested words list.

    G{classtree StopList}
    """

    def __init__(self, stopListFile):
        """Extract every words of the file and store them in a list.

        The file must contains one word per line. The method dosen't lower any
        word.

        @param stopListFile:
            Path to the stoplist file.
        @type stopListFile: str
        """
        self.stopListFile = stopListFile
        self.words = []
        if self.stopListFile:
            try:
                with open(self.stopListFile) as f:
                    self.words = [x.strip('\n') for x in f.readlines()]
            except IOError:
                lg.error('Cannot open file "%s". Stop list won\'t be used'
                         % (self.stopListFile))
        self.size = len(self.words)

    def add_words(self, wordList):
        """Add words to the stoplist.

        @param wordList:
            The words to add.
        @type wordList: list
        """
        for word in wordList:
            self.add_word(word)

    def add_word(self, word):
        """Add a word to the stoplist.

        @param word:
            The word to add.
        @type word: str
        """
        self.words.append(word)
        self.size += 1

    def remove_words(self, wordList):
        """Remove words from the stoplist.

        @param wordList:
            The words to remove.
        @type wordList: list
        """
        for word in wordList:
            self.remove_word(word)

    def remove_word(self, word):
        """Remove a word from the stoplist.

        @param word:
            The word to remove.
        @type word: str
        """
        try:
            self.words.remove(word)
            self.size -= 1
        except ValueError:
            print('WARNING: "' + word + '" is not in stop list')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Classes for context monitoring and context change detection.

The context is determined by the input buffers contained in the Callback class.
It is necessary to monitor the context in order to know what the user is typing
and compute accurate predictive suggestions.
"""

from tipy.char import blankspaces, separators, is_word_char, last_word_char
from tipy.tknz import ReverseTokenizer, ForwardTokenizer
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


class ContextChangeDetector(object):
    """Detect context change.

    A context change can occure when some special characters appear in the
    buffers:
        - Word characters indicate that the current token is a (partial) word.
        - Blankspaces indicate the separations between two words. They mark
          the end of the current token.
        - Separators indicate a separation between two words. The dot separator
          mark the end of the sentence (and of the token) so the next token and
          suggestions should begin with an uppercase letter.
        - Special characters are non-printable characters such as backspace and
          arrow keys which are used to modify the input buffers.

    It is important to detect context change because some operations such as
    n-gram learning from input or suggested words suppression have to be carried
    out upon context changes.

    G{classtree ContextChangeDetector}
    """

    def __init__(self, lowercase, config):
        """ContextChangeDetector creator.

        @param config:
            It is used to retrieve the ContextMonitor settings from the
            configuration file.
        @type config: L{drvr.Configuration}
        @param lowercase:
            Indicate if the tokens should be convert to lowercase.
        @type lowercase: boolean
        """
        self.lowercase = lowercase
        self.config = config
        self.monitoredScopeSize = self.config.getas(
            'ContextMonitor', 'monitored_scope', 'int')
        self.monitoredScope = ''

    def update_monitored_scope(self, string):
        """Move the monitored scope according to the string length.

        @param string:
            Every characters inputed in the monitored buffer.
        @type string: str
        """
        if len(string) <= self.monitoredScopeSize:
            self.monitoredScope = string
        else:
            self.monitoredScope = string[:-self.monitoredScopeSize]

    def context_change(self, leftBuffer):
        """Check if the context has changed.

        To determine if a context change occure or not it is important to
        scan the input left buffer and the monitored scope. A change occure if:
            - The monitored scope is not part of the left buffer.
            - The monitored scope is part of the left buffer and a separator
              character appear in the left buffer part wich is not the monitored
              scope.

        @param leftBuffer:
            The input left buffer.
        @type leftBuffer: str

        @return:
            True or False weither the context has changed or not.
        @rtype: boolean
        """
        prevContext = self.monitoredScope
        currContext = leftBuffer
        if len(prevContext) == 0:
            if len(currContext) == 0:
                return False
            else:
                return True
        iIdx = currContext.rfind(prevContext)
        if iIdx == -1:
            return True
        rest = currContext[iIdx + len(prevContext):]
        idx = last_word_char(rest)
        if idx == -1:
            if len(rest) == 0:
                return False
            last_char = currContext[iIdx + len(prevContext) - 1]
            if is_word_char(last_char):
                return True
            else:
                return False
        if idx == len(rest) - 1:
            return False
        return True

    def change(self, leftBuffer):
        """Return the (part of the) token(s) appearing after a change.

        When a change occure it is necessary to retrieve the characters forming
        (partial) tokens which have been inputed AFTER the change and this is
        what this method do.
        Weither a change occure or not is determined by self.context_change().

        @note: If no change have been registered yet then the leftBuffer is
               returned.

        @param leftBuffer:
            The input left buffer.
        @type leftBuffer: str

        @return:
            (Part of) tokens inputed after the last change.
        @rtype: list
        """
        prevContext = self.monitoredScope
        currContext = leftBuffer
        if len(prevContext) == 0:
            return currContext
        iIdx = currContext.rfind(prevContext)
        if iIdx == -1:
            return currContext
        result = currContext[iIdx + len(prevContext):]
        if self.context_change(leftBuffer):
            tokenizer = ReverseTokenizer(prevContext, self.lowercase)
            firstToken = tokenizer.next_token()
            if not len(firstToken) == 0:
                result = firstToken + result
        return result


class ContextMonitor(object):   # observer.Observer
    """Monitire user current context.

    This class monitore the input buffers in order to:
        - Tokenize the input and use the tokens for prediction.
        - Identify context changes.

    G{classtree ContextMonitor}
    """

    def __init__(self, config, predictorRegistry, callback):
        """ContextMonitor creator.

        @param config:
            It is used to retrieve the ContextMonitor settings from the
            configuration file.
        @type config: L{drvr.Configuration}
        @param predictorRegistry:
            It is used to access the predictors's learn() methods. Also, the
            ContextMonitor is used by the predictors to access the input
            buffers.
        @type predictorRegistry: L{PredictorRegistry}
        @param callback:
            As the callback hold the input buffers and the ContextMonitor
            operate on these buffers, it is used to access the input buffers
            from inside the ContextMonitor.
        @type callback: L{Callback}
        """
        self.config = config
        self.lowercase = self.config.getas(
            'ContextMonitor', 'lowercase', 'bool')
        self.liveLearning = self.config.getas(
            'ContextMonitor', 'live_learning', 'bool')
        self.predictorRegistry = predictorRegistry
        self.callback = callback
        self.contextChangeDetector = ContextChangeDetector(
            self.lowercase, self.config)
        self.predictorRegistry.contextMonitor = self

    def context_change(self):
        """Check if a context change occure.

        @return:
            Return True or False weither a context change occure.
        @rtype: bool
        """
        return self.contextChangeDetector.context_change(self.left_buffer())

    def update(self):
        """Check if context changes occure and learn what need to be learnt.

        This method is called by Driver.predict() after the predictions have
        been computed. It check if a context change occure in the input
        buffers and if so, it learn the words that need to be learnt if the
        predictor's learning mode is ON. Finaly, it update the monitored scope.
        """
        change = self.contextChangeDetector.change(self.left_buffer())
        if self.liveLearning and change:
            self.learn(change)
        self.contextChangeDetector.update_monitored_scope(self.left_buffer())

    def learn(self, string):
        """Learn n-grams from the input buffers.

        Trigger the learn() method of each predictor of the registry. This
        method use the input buffers to create n-grams and add them to the
        predictors's databases or memory so that the program learn from the
        user input.

        @param string:
            The string to learn.
        @type string: str
        """
        tokens = []
        tok = ForwardTokenizer(string, self.lowercase, blankspaces, separators)
        while tok.has_more_tokens():
            token = tok.next_token()
            tokens.append(token)
        if tokens:
            tokens = tokens[:-1]
        for predictor in self.predictorRegistry:
            predictor.learn(tokens)

    def prefix(self):
        """Return the token just before the cursor.

        @return:
            The token just before the cursor or an empty string if there is
            none.
        @rtype: str
        """
        return self.left_token(0)

    def suffix(self):
        """Return the token just after the cursor.

        @return:
            The token just after the cursor or the empty string if there is
            none.
        @rtype: str
        """
        return self.right_token(0)

    def left_token(self, index):
        """Return the token at a given index in the left input buffer.

        @param index:
            The index of the token to retrieve in the left input buffer.
        @type index: int

        @return:
            The token at index 'index' in the left input buffer or an empty
            string if the token dosen't exists.
        @rtype: str
        """
        leftInput = self.left_buffer()
        tok = ReverseTokenizer(leftInput, self.lowercase)
        i = 0
        while tok.has_more_tokens() and i <= index:
            token = tok.next_token()
            i += 1
        if i <= index:
            token = ''
        return token

    def right_token(self, index):
        """Return the token at a given index in the right input buffer.

        @param index:
            The index of the token to retrieve in the right input buffer.
        @type index: int

        @return:
            The token at index 'index' in the right input buffer or an empty
            string if the token dosen't exists.
        @rtype: str
        """
        tok = ForwardTokenizer(self.right, self.lowercase)
        i = 0
        while tok.has_more_tokens() and i <= index:
            token = tok.next_token()
            i += 1
        if i <= index:
            token = ''
        return token

    def previous_tokens(self, index, change):
        """Return the token just before the change token (if any).

        This method is called in some predictors's learn() method. It retrieve
        the token that appear just before the change token and has already
        been learnt before (or should have). The previous token is used to fill
        the n-grams.

        @param index:
            Index of the previous token.
        @type index: int
        @param change:
            The change token.
        @type change: str

        @return:
            The token just before the change token or an empty string if there
            is none.
        @rtype: str
        """
        return self.left_token(index + len(change))

    def left_buffer(self):
        """Use the callback to get the value of the left buffer.

        @return:
            The left input buffer.
        @rtype: str
        """
        return self.callback.left

    def right_buffer(self):
        """Use the callback to get the value of the right buffer.

        @return:
            The right input buffer.
        @rtype:
            str
        """
        return self.callback.right

    def make_completion(self, suggestion):
        """Compute the completion string given a suggested word.

        This method compute and return the completion string using the token
        just before the cursor (prefix) and the suggested word (suggestion).
        The suggestion should be the word that the user choose from the
        suggested words list.

        For instance, if the prefix is::
            "wor"
        And the suggestion is::
            "world"
        Then this method will compute the completion::
            "ld"

        If the character before the cursor is a blankspace or a separator then
        the prefix should be empty::
            ""
        Then if the suggestion is::
            "guilty"
        This method will compute the completion::
            "guilty"

        If the suggestion and the prefix don't match then False is returned.
        This should never happen as suggestions completing an input word should
        always match it. Still, I prefer to check it at the cost of some lower()
        and startswith() calls.

        @param suggestion:
            The suggested word from which to compute the completion.
        @type suggestion: str
        """
        prefix = self.prefix()
        if suggestion.lower().startswith(prefix.lower()):
            return suggestion[len(prefix):]
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Some variables and functions to help handling characters categories."""


from unicodedata import category


blankspaces = ' \f\n\c\r\t\v'
"""@var: A string containing every blankspaces characters."""

separators = '`~!@#$%^&*()_-+=\\|]}[{\";:/?.>,<†„“।॥ו–´’‘‚י0123456789ः'
"""@var: A string containing every separators characters."""


def first_word_char(string):
    """Return the index of the first word character in a string.

    @return:
        The index of the first word character in the string 'string' or -1 if
        the string contains no word character.
    @rtype: int
    """
    for i, ch in enumerate(string):
        if is_word_char(ch):
            return i
    return -1


def last_word_char(string):
    """Return the index of the last word character in a string.

    @return:
        The index of the last word character in the string 'string' or -1 if
        the string contains no word character.
    @rtype: int
    """
    result = first_word_char(string[::-1])
    if result == -1:
        return -1
    return len(string) - result - 1


def is_word_char(char):
    """Check if a character is a word character.

    @return:
        True or False weither the character 'char' is a word character, i.e. a
        printable non-ponctuation character.
    @rtype: bool
    """
    if category(char)[0] == "L":
        return True
    return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Callback class holds the buffers and allow the outer world to access them."""


class Callback(object):
    """Allow access to the current buffers from anywhere.

    This class is used as a callback to retrieve and modify the input buffers.
    There is two input buffers the left buffer and the right buffer:
        - Left buffer contains every characters preceding the cursor.
        - Right buffer contains every characters following the cursor.

    It is possible to subclass this class and override its methods in order to
    use the program in other GUIs or command line interfaces.

    G{classtree Callback}
    """

    def __init__(self):
        """Initialize the left and right buffer to empty string."""
        self.left = ''
        self.right = ''

    def update(self, character, offset=0):
        """Update the left and right buffers.

        This method should be called after each character addition or deletion
        or after each cursor move.
        In the default GUI this method is called when a signal indicating that
        the input text have changed or that the cursor position have changed.

        @param character:
            The character which has been inputted. It could be any character
            associated to keyboard keys including backspace, arrow keys...
        @type character: str
        @param offset:
            Indicate the cursor position offset against the previous cursor
            position. 0 means no move, a positive value means move to the right
            and a negative value means move to the left.
        @type offset: int
        """
        if character == '\b' and len(self.left) >= offset:
            self.left = self.left[:-offset]
        elif character == '\x1B[D' and len(self.left) >= offset:
            self.right = self.left[offset:] + self.right
            self.left = self.left[:offset]
        elif character == '\x1B[C' and len(self.right) >= offset:
            self.left += self.right[:offset]
            self.right = self.right[offset:]
        else:
            self.left += character



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Selector select suggested words in the list depending on various criteria.
"""

from collections import OrderedDict


class Selector(object):
    """The Selector class select the best suggested words among all.

    The goal of the selector is to clean the suggested words list (contained in
    a Prediction instance which should have been returned by a
    L{Merger<mrgr>} L{merge()<mrgr.Merger.merge>} method and remove
    the words which are too much (the configuration specify the excpected number
    of suggested  words).

    G{classtree Selector}
    """

    def __init__(self, config, contextMonitor):
        """Selector creator.

        @param config:
            The configuration dictionary is used in order to retrieve the
            Selector settings from the config file.
        @type config: L{drvr.Configuration}
        @param contextMonitor:
            The ContextMonitor is used to check if a context change occure. In
            which case the suggested words list must be cleared because
            prediction fully depends on the context.
        @type contextMonitor: L{ContextMonitor}
        """
        self.contextMonitor = contextMonitor
        self.config = config
        self.suggestions = self.config.getas('Selector', 'suggestions', 'int')
        self.suggestedWords = []

    def select(self, prediction):
        """Select suggested words in the suggestions list.

        Selecting suggested words consists in:
            - List every suggestions words.
            - Remove duplicate suggested words.
            - Shorten the suggested words list so that it contains the desired
              number og suggested words.

        @param prediction:
            The list of suggestions from which to carry out the selection.
        @type prediction: L{Prediction}
        """
        words = [sugg.word for sugg in prediction]
        if self.contextMonitor.context_change():
            self.suggestedWords = []
        self.rm_duplicate(words)
        if len(words) > self.suggestions:
            words = words[:self.suggestions]
        self.suggestedWords += words
        return words

    def rm_duplicate(self, words):
        """Remove duplicate words in the list and keep the original order.

        @param words:
            The words to filter.
        @type words: list
        """
        self.suggestedWords = list(
            OrderedDict.fromkeys(self.suggestedWords + words))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""The main file parse the command line and run the program in the desired way.

There's three possible runing options:
    - -c or --cmdl: Run the program using the command line interface (not
        really suitable)
    - -g or --gui: Run the program using the graphic user interface.
    - -p or --profile: Simulate the predictions computation for a sample
        string and print the profiling statistics for the session. This is
        useful to see what part of the program took the largest part of the
        program's execution time.

Examples of profiling use::
    tipy -p '' -l 10
        -> Simulate the input and predction of ten consecutive inputs of the
           'hello world' string (default string).

    tipy -p 'the seasons wither'
        -> Simulate the input and predction of the 'the seasons wither'
           string.

@note: You must specify your own test string after -p or --profile, if
    none are specified a default one will be used.
@note: The option -l or -loop indicate how many times the profiling
    operation have to be carried out.

@todo 0.1.0:
    Add memoize classes and decorators to improve performance.
"""

from PyQt5 import QtWidgets
from tipy.clbk import Callback
from tipy.drvr import Driver
from tipy.gui import MainWindow
from tipy.lg import lg
from sys import argv, exit
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
from getopt import getopt, GetoptError
from cProfile import Profile, runctx
from pstats import Stats
from os import path, environ


CONFIG_FILE = ''

def set_config_file():
    """Look for the best directory where a config file can be."""
    global CONFIG_FILE
    locations = [
        path.join(path.expanduser("~"), '.config/tipy'),
        path.join(path.dirname(path.realpath(__file__)), 'cfg'),
        "/usr/share/tipy",
        environ.get("TIPY_CONFIG_PATH"),
    ]
    for loc in locations:
        try:
            if path.isfile(path.join(loc, "tipy.ini")):
                CONFIG_FILE = path.join(loc, "tipy.ini")
                lg.info('Using config file "%s/tipy.ini"' % loc)
                break
        except IOError:
            pass
        except TypeError:
            pass


def run_profile(string='hello world', profileLoops=1):
    """Initialize the Driver instance and start a cmdl REP loop."""
    for i in range(profileLoops):
        callback = Callback()
        driver = Driver(callback, CONFIG_FILE)
        for c in string:
            callback.update(c)
            suggestions = driver.predict()
            for p in suggestions:
                print(p)
            print('-- Context: %s<|>' % (driver.callback.left))


def run_cmdl():
    """Initialize the Driver instance and start a cmdl REP loop."""
    callback = Callback()
    driver = Driver(callback, CONFIG_FILE)
    while True:
        buffer = input('> ')
        callback.update(buffer)
        suggestions = driver.predict()
        for p in suggestions:
            print(p)
        print('-- Context: %s<|>' % (driver.callback.left))


def run_gui():
    """Initialize the Driver instance and run the """
    callback = Callback()
    driver = Driver(callback, CONFIG_FILE)
    app = QtWidgets.QApplication(argv)
    mySW = MainWindow(driver)
    mySW.show()
    exit(app.exec_())


def usage():
    """Print the usage message."""
    print('USAGE:   python3 main.py [-h -c -g -p <string> -l <loops>]\n')
    print('OPTIONS:')
    print('\t-h or --help: print the help')
    print('\t-c or --cmdl: use the command line version')
    print('\t-g or --gui: use the graphic user interface version')
    print('\t-p or --profile: use the profile module to evaluate program '
          'performance.')
    print('\t\t<string> is the string to use for prediction computation.')
    print('\t-l or --loops: run the profiling operation multiple times.')
    print('\t\t<loops> is the number of loops.')


def main():
    """The main function of the program.

    Take care of the command line options and run the program (gui or command
    line) or print the help.
    """
    try:
        opts, args = getopt(
            argv[1:], "hcgp:l:",
            ['help', 'cmdl', 'gui', 'profile=', 'loop='])
    except GetoptError as err:
        print(str(err))
        usage()
        exit(2)
    mode = 'gui'
    profileString = 'hello world'
    profileLoops = 1
    for o, a in opts:
        if o in ("-h", "--help"):
            usage()
            exit()
        elif o in ("-c", "--cmdl"):
            mode = 'cmdl'
        elif o in ("-g", "--gui"):
            mode = 'gui'
        elif o in ("-p", "--profile"):
            mode = 'profile'
            if a:
                profileString = a
        elif o in ("-l", "--loop"):
            if a:
                profileLoops = int(a)
        else:
            assert False, "ERROR: Unrecognized option"
    set_config_file()
    if mode == 'cmdl':
        run_cmdl()
    elif mode == 'profile':
        pr = Profile()
        pr.enable()
        runctx('run_profile(profileString, profileLoops)', {'profileString': profileString, 'run_profile': run_profile, 'profileLoops': profileLoops}, locals())
        pr.disable()
        s = StringIO()
        ps = Stats(pr, stream=s).sort_stats('time')
        ps.print_stats()
        print(s.getvalue())
    else:
        run_gui()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""The Driver class holds everything the program needs to perform prediction.

It gather instances of needed classes, implement some wrapper methods and the
very important L{drvr.Driver.predict} method which actualy compute the suggested
words.
"""

try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import ConfigParser
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
from tipy.clbk import Callback
from tipy.slct import Selector
from tipy.prdct import PredictorRegistry, PredictorActivator
from tipy.cntxt import ContextMonitor
from os import path
from tipy.lg import lg


class UnknownTypeCastError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr('Unknown cast type "%s"' % self.value)


class MissingConfigurationSection(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr('Section "%s" is missing from the configuration'
                    % self.value)


class MissingConfigurationOption(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr('Section "%s" is missing from the configuration'
                    % self.value)


class Configuration(ConfigParser):
    """Copy a configuration file (ini format) in memory.

    This class subclass the Configparser class. Configparser is used to read() a
    configuration file (ini format) in memory in the form of a dictionary
    associating sections and options.
    This class implement a new method which allow to retrieve and cast a
    configuration option and asserts the option do exists and can be casted.
    The config file could be edited by the user and you know... never trust
    user input.

    G{classtree Configuration}
    """

    def __init__(self):
        """Configuration creator."""
        super().__init__()

    def getas(self, section, option, typeCast=None):
        """A more secure way to retrieve configuration options.

        This method check if the section and the option is in the configuration
        dictionary, else it raise an error. Also this method allow an optional
        parameter for casting the result before returning it. Allowed type
        casting are:
            - bool: cast to bool
            - int: cas to int
            - float: cast to float
            - list: cast to list
            - intlist: cast to list and cast each element to int
            - floatlist: cast to list and cast each element to float
        The method makes sure the casting is possible.

        @param section:
            The section from which to retrieve the option.
        @type section: str
        @param option:
            The option to retrieve inside the section.
        @type option: str
        @param typeCast:
            Indicate how to cast the result. If no value are passed the
            result is returned as a string.

        @return:
            The casted value of the given option found inside the given section.
        @rtype:
            str or int or bool or float or list (depends on "typeCast" param)

        @raise MissingConfigurationSection:
            If the given section cannot be found in the configuration
            dictionary.
        @raise MissingConfigurationOption:
            If the given option cannot be found in the configuration dictionary.
        @raise UnknownTypeCastError:
            If the given typeCast value is not "bool", "int", "float", "list",
            "intlist" or "floatlist".
        """
        if not self.has_section(section):
            raise MissingConfigurationSection(section)
        if not self.has_option(section, option):
            raise MissingConfigurationOption(option)
        if not typeCast or typeCast == 'str':
            return self.get(section, option)
        if typeCast == 'bool':
            return self.getboolean(section, option)
        if typeCast == 'int':
            return self.getint(section, option)
        if typeCast == 'float':
            return self.getfloat(section, option)
        if typeCast == 'list':
            return self.get(section, option).split()
        if typeCast == 'intlist':
            return [int(x) for x in self.get(section, option).split()]
        if typeCast == 'floatlist':
            return [float(x) for x in self.get(section, option).split()]
        raise UnknownTypeCastError(typeCast)


class Driver:
    """The Driver class gather classes inctances and variables of the program.

    G{classtree Driver}
    """

    def __init__(self, callback, configFile=''):
        """The driver class. It hold every elements needed for the prediction.

        @param callback:
            The callback is used to access the input buffers from anywhere.
        @type callback: L{Callback}
        @param configFile:
            Path of the configuration file.
        @type configFile: str
        """
        self.configFile = configFile
        self.configuration = self.make_config()
        self.callback = callback
        self.predictorRegistry = PredictorRegistry(self.configuration)
        self.contextMonitor = ContextMonitor(
            self.configuration, self.predictorRegistry, callback)
        self.predictorActivator = PredictorActivator(
            self.configuration, self.predictorRegistry)
        self.selector = Selector(self.configuration, self.contextMonitor)

    def predict(self):
        """Request suggested words to predictors.

        This method:
            - Do the next two steps until it cannot get more suggestions.
            - Call the L{prdct.PredictorActivator.predict} which:
                - Call the L{prdct.Predictor.predict} method of each
                  predictors in the predictorRegistry. Each predict() method
                  should return a Prediction instance containing the suggested
                  words computed by the predictor (it may be empty).
                - Merge the Prediction instances into a single Prediction
                  instance.
            - Select the best suggestions in the Prediction instance and remove
              the excess.
            - Learn from what the user have typped.
            - Return the selected suggestions.

        @return:
            The suggested words list.
        @rtype: list
        """
        factor = 1
        predictions = self.predictorActivator.predict(factor)
        result = self.selector.select(predictions)

        previousPredictions = predictions
        while len(result) < self.selector.suggestions:
            predictions = self.predictorActivator.predict(factor)
            if len(predictions) > len(previousPredictions):
                factor += 1
                result = self.selector.select(predictions)
                previousPredictions = predictions
            else:
                lg.warning('WARNING: Expected number of suggestions cannot be '
                           'reached.')
                break
        self.learn_from_buffers()
        return result

    def learn_from_buffers(self):
        """Simple ContextMonitor.update() wrapper for comprehension sake."""
        self.contextMonitor.update()

    def make_completion(self, suggestion):
        """Simple ContextMonitor.make_completion() wrapper."""
        return self.contextMonitor.make_completion(suggestion)

    def close_databases(self):
        """Close every opened predictors database."""
        self.predictorRegistry.close_database()

    def make_config(self):
        """Initialize the config dictionary.

        This method first try to read the configuration file and parse it into
        a Configuration instance (dictionary).
        If the config file is empty or dosen't exists, than a default config
        dictionary is created.

        @return:
            The Configuration instance holding every settings (dictionary
            style).
        @rtype: L{drvr.Configuration}
        """
        config = Configuration()
        if config.read(self.configFile) == []:
            lg.warning('Cannot open config file "%s"' % self.configFile)
            config.readfp(StringIO(
                """
                [Global]
                language = en

                [GUI]
                font_size = 10

                [MinerRegistry]
                miners = CorpusMiner FbMiner 

                [CorpusMiner]
                class = CorpusMiner
                texts = ../txt/brown.txt 
                database = ../databases/corp.db
                lowercase = False
                n = 3

                [DictMiner]
                class = DictMiner
                dictionary = /usr/share/dict/words
                database = ../databases/dict.db

                [FbMiner]
                class = FacebookMiner
                accesstoken = CAAUh40uO4aIBACsN37erjPUfrs0IdRZAxpbc7IhGlWpKX1REif4NPhqqD4tdZCdZBgT1J3K41ZCB3CHsjcWV78tNEDtwPwmHZBPJ3NgLV29rZCSl4vkGEqPdyEq9UNRR7OkMS1195LWO7W5jUjBrsuRYZBT7IemlZAg4FHZANrZBkaSZCwzxyrbplNYSYRB28vIdzpeLSh2s0gFZAykLoaeDQ6IKk02K9UCnK9QZD
                database = ../databases/fb.db
                lowercase = False
                n = 3
                last_update = 1433879088

                [TwitterMiner]

                [PredictorRegistry]
                predictors = CorpusNgramPredictor InputNgramPredictor DictionaryPredictor

                [ContextMonitor]
                live_learning = True
                monitored_scope = 80
                lowercase = True

                [Selector]
                suggestions = 6
                greedy_suggestion_threshold = 0

                [PredictorActivator]
                predict_time = 1000
                max_partial_prediction_size = 50
                merging_method = probabilistic
                stoplist = ../stoplists/insanities_en.stoplist

                [CorpusNgramPredictor]
                class = WeightNgramPredictor
                database = ../databases/corp.db
                deltas = 0.01 0.1 0.89 
                learn = False

                [InputNgramPredictor]
                class = WeightNgramPredictor
                database = ../databases/user.db
                deltas = 0.01 0.1 0.89 
                learn = True

                [FbNgramPredictor]
                class = WeightNgramPredictor
                database = ../databases/fb.db
                deltas = 0.01 0.1 0.89 
                learn = False

                [LateOccurPredictor]
                class = LastOccurPredictor
                lambda = 1
                n_0 = 1
                cutoff_threshold = 20

                [MemorizePredictor]
                class = MemorizePredictor
                memory = ../txt/memory.txt
                trigger = 3
                learn = True

                [DictionaryPredictor]
                class = DictionaryPredictor
                database = ../databases/dict.db
                probability = 0.000001
                """))
        else:
            for section in config._sections:
                for key, value in config._sections[section].items():
                    config.set(section, key, value.replace(
                        '~', path.expanduser("~")))
        return config



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Merger classes to merge results from several predictors."""

import tipy.prdct
from abc import ABCMeta, abstractmethod

class Merger(object):
    """Base class for all Mergers.

    G{classtree Merger}
    """

    __metaclass__ = ABCMeta
    MIN_PROBABILITY = 0.0
    MAX_PROBABILITY = 1.0

    def __init__(self):
        """Merger creator."""
        self.name = "Merger dosen't set any name"

    def filter(self, prediction):
        """Sort the tokens according to their probabilities.

        The duplicate tokens are merged to a single token (their probabilities
        are summed).

        @param prediction:
            It is a list of Suggestion instances.
        @type prediction: L{prdct.Prediction}

        @return:
            The sorted Prediction instance.
        @rtype: Prediction
        """
        seen_tokens = set()
        result = tipy.prdct.Prediction()
        for i, suggestion in enumerate(prediction):
            token = suggestion.word
            if token not in seen_tokens:
                for j in range(i + 1, len(prediction)):
                    if token == prediction[j].word:
                        suggestion.probability += prediction[j].probability
                        if suggestion.probability > self.MAX_PROBABILITY:
                            suggestion.probability = self.MAX_PROBABILITY
                        elif suggestion.probability < self.MIN_PROBABILITY:
                            suggestion.probability = self.MIN_PROBABILITY
                seen_tokens.add(token)
                result.add_suggestion(suggestion)
        return result

    @abstractmethod
    def merge(self, predictionList):
        """Method for merging predictors's prediction list.

        The merge() method is the main method of the Merger. It must be
        implemented by every Mergers.
        """
        raise NotImplementedError("Method must be implemented")


class ProbabilisticMerger(Merger):
    """Simple Merger which merge suggestions based on their probabilities.

    This Merger does not modify the suggestions probabilities. Thus, they are
    sorted in descending order according to their probabilities.

    G{classtree ProbabilisticMerger}
    """
    def __init__(self):
        """ProbabilisticMerger creator."""
        super(self.__class__, self).__init__()
        self.name = 'ProbabilisticMerger'

    def merge(self, predictionList):
        """merge the suggestions in a single list and sort them.

        @param predictionList:
            The list of Prediction instances. Each predictors return a
            Prediction instance which is added to the list by the
            PredictorActivator.
        @type predictionList: list

        @return:
            The merged Prediction instance containing every suggestions of
            every Prediction instances sorted in descending order according to
            their probabilities.
        @rtype: L{prdct.Prediction}
        """
        result = tipy.prdct.Prediction()
        for prediction in predictionList:
            for suggestion in prediction:
                result.add_suggestion(suggestion)
        return(self.filter(result))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Classes for predictors and to handle suggestions and predictions."""

from sys import exit
from tipy.db import SqliteDatabaseConnector
from tipy.mrgr import ProbabilisticMerger
from abc import ABCMeta, abstractmethod
from tipy.stpl import *
from math import exp
from multiprocessing import Queue, Process


class Suggestion(object):
    """A suggestion consists of a string and a probility.

    The string is in fact a token and the probability is the probability
    associated to this token by the predictor which compute it. The
    probability reflect how accurate the word is over prediction.

    G{classtree Suggestion}
    """

    def __init__(self, word, probability):
        """Suggestion creator

        A suggestion is a couple formed by a suggested word and its probability.

        @note: the probabilities of each predictors should have the same
            weight. Otherwise the suggestion selection will be truncated.

        @param word:
            The suggested word.
        @type word: str
        @param probability:
            The suggested word probability. It is compute by the predictors.
        @type probability: float
        """
        self.word = word
        self.probability = probability

    def __eq__(self, other):
        """Override the == operator in order to compare instances equality.

        Two Suggestion instances are equal if their word and probability are
        equal.

        @param other:
            The Suggestion instance to compare to this one (self).
        @type other: L{Suggestion}

        @return:
            True if the two instances are equal, False otherwise.
        @rtype:
            bool
        """
        if self.word == other.word and self.probability == other.probability:
            return True
        return False

    def __lt__(self, other):
        """Override the < operator in order to compare instances.

        A Suggestion instance is less than another if its probability is less
        than the probability of the other. If their probabilities are equal then
        the Suggestion instance is less than the other if its word is
        alphabetically 'before' the word of the other instance.

        @param other:
            The Suggestion instance to compare to this one (self).
        @type other: L{Suggestion}

        @return:
            True if the this instance (self) is less than the other one, False
            otherwise.
        @rtype: bool
        """
        if self.probability < other.probability:
            return True
        if self.probability == other.probability:
            return self.word < other.word
        return False

    def __gt__(self, other):
        """Override the > operator in order to compare instances.

        A Suggestion instance is greater than another if its probability is
        greater than the probability of the other. If their probabilities are
        equal then the Suggestion instance is greater than the other if its word
        is alphabetically 'after' the word of the other instance.

        @param other:
            The Suggestion instance to compare to this one (self).
        @type other: L{Suggestion}

        @return:
            True if the this instance (self) is greater than the other one,
            False otherwise.
        @rtype: bool
        """
        if self.probability > other.probability:
            return True
        if self.probability == other.probability:
            return self.word > other.word
        return False


class Prediction(list):
    """Class for predictions from predictors.

    A Prediction instance hold multiple Suggestion instances. It is a list of
    Suggestion instances and the list is kept ordered according to the
    suggestions probabilities.
    Every predictors should return a Preidiction instance.

    G{classtree Prediction}
    """

    def __init__(self):
        """Prediction creator."""
        pass

    def __eq__(self, other):
        """Override the == operator in order to compare instances.

        Two Prediction instances are equal if they contain the same items
        (thus, have the same length).
        """
        if self is other:
            return True
        if len(self) != len(other):
            return False
        for i, s in enumerate(other):
            if not s == self[i]:
                return False
        return True

    def add_suggestion(self, suggestion):
        """Add a suggestion in the Prediction list.

        The suggestion is added at the correct index so that the Prediction
        list remains ordered.

        @note: Using insert() and a while loop seams a little faster than using
               sorted(). Also, using insort from the bisect module seems to 
               produce similar benchmarks.
        """
        if len(self) == 0:
            self.append(suggestion)
        else:
            i = 0
            while i < len(self) and suggestion < self[i]:
                i += 1
            self.insert(i, suggestion)


class PredictorActivator(object):
    """Query the predictors listed in the registry to get their suggestions.

    This class has access to a PredictorRegistry and asks the predictors listed
    in this PredictorRegistry to call their predict() method, store the
    resulting Prediction instances, merge them into a single Prediction
    instance and return it.

    G{classtree PredictorActivator}
    """

    def __init__(self, config, predictorRegistry):
        """PredictorActivator creator.

        @param config:
            The configuration dictionary is used in order to retrieve the
            PredictorActivator settings from the config file.
        @type config: L{drvr.Configuration}
        @param predictorRegistry:
            The class needs to access the PredictorRegistry to call their
            predict() method.
        @type predictorRegistry: L{PredictorRegistry}
        """
        self.config = config
        self.predictorRegistry = predictorRegistry
        self.predictionList = []
        self.maxPartialPredictionSize = self.config.getas(
            'PredictorActivator', 'max_partial_prediction_size', 'int') + 1
        self.mergingMethod = self.config.getas(
            'PredictorActivator', 'merging_method')
        self.stopListFile = self.config.getas(
            'PredictorActivator', 'stoplist')
        self.stopList = StopList(self.stopListFile)
        if self.mergingMethod.lower() == "probabilistic":
            self.merger = ProbabilisticMerger()
        else:
            lg.error('Unknown merging method')
            exit(1)

    def pred_worker(self, predictor, queue, factor):
        """Worker function for the predictor predict() methods.

        This method is used as the predictors workers target. It push the
        predictor's L{prdct.Predictor.predict} method result (a
        L{Prediction} instance) in a queue (which is used because it is
        thread-safe).

        @param predictor:
            The Predictor based class instance.
        @type predictor: L{Predictor} based class.
        @param queue:
            A queue in which the result will be pushed.
        @type queue:
            multiprocessing.Queue
        @param factor:
            A factor used to increase the number of suggestions.
        @type factor: int
        """
        queue.put(predictor.predict(
                self.maxPartialPredictionSize * factor, self.stopList.words))

    def predict(self, factor=1):
        """Build a list of every predicted words.

        Call the predict() method of every predictors in the registry then
        merge their Prediction into a single Prediction instance.

        @change:
            - 16/06/15: The method now uses multi-processing. It concurrently
                runs every predictors's predict() method which allow a
                significant speed augmentation. The queue is used because it is
                thread safe. The point is that when the threads args are passed
                to the L{PredictorActivator.pred_worker()}, they are
                packed up with pickle, shipped to the other process, where they
                are unpacked used. A list wouldn't be passed but would be
                cloned.

        @note:
            Using multi-processing allow significant speed boost. The next
            benchmark have been maid runing 100 * 10 different contexts
            predictions::

            Total time without multi-processing: 86.785 s
            Total time wit multi-processing:     76.513 s

        @todo 0.0.9:
            Demonize the processes, set a timeout value. When the time runs out
            the unfinished workers return their results as is. This can alter
            the prediction qality but totaly avoid any possible "slow 
            predictions".

        @param factor:
            A factor used to increase the number of suggestions.
        @type factor: int

        @return:
            The merged Prediction instance containing every suggestions of
            every Prediction instances sorted in descending order according to
            their probabilities.
        @rtype: L{Prediction}
        """
        self.predictionList[:] = []
        jobs = []
        queue = Queue()
        for predictor in self.predictorRegistry:
            p = Process(
                target=self.pred_worker, args=(predictor, queue, factor,))
            jobs.append(p)
            p.start()
        for job in jobs:
            job.join()
        for x in range(len(jobs)):
            self.predictionList.append(queue.get())
        return self.merger.merge(self.predictionList)


class PredictorRegistry(list):  # observer.Observer,
    """List every predictors instances that are to be used for word prediction.

    G{classtree PredictorRegistry}
    """

    def __init__(self, config):
        """PredictorRegistry creator.

        @param config:
            config is used to retrieve the PredictorRegistry settings and each
            Predictor settings from the config file. Also it needs to be passed
            to the predictors instances to allow them to retrieve their settings
            from the config file too.
        @type config: L{drvr.Configuration}
        """
        self._contextMonitor = None
        self.config = config
        self.contextMonitor = None

    def contextMonitor():

        def fget(self):
            return self._contextMonitor

        def fset(self, value):
            if self._contextMonitor is not value:
                self._contextMonitor = value
                self[:] = []
                self.set_predictors()

        def fdel(self):
            del self._contextMonitor

        return locals()
    contextMonitor = property(**contextMonitor())

    def set_predictors(self):
        """Read the configuration file and create needed predictors."""
        if self.contextMonitor:
            self[:] = []
            preds = self.config.getas('PredictorRegistry', 'predictors', 'list')
            for predictor in preds:
                self.add_predictor(predictor)

    def add_predictor(self, predictorName):
        """Create and add a predictor to the list.

        Create a predictor instance according to the predictor name and add
        it to the list.

        @param predictorName:
            The name of the predictor. It is used to retrieve the predictor
            settings from the config. It must correspond to a section of the
            config, otherwise no predictor will be created and added.
        @type predictorName: str
        """
        predictorClass = self.config.getas(predictorName, 'class')
        if predictorClass == 'WeightNgramPredictor':
            predictor = WeightNgramPredictor(
                self.config, self.contextMonitor, predictorName)
        elif predictorClass == 'LastOccurPredictor':
            predictor = LastOccurPredictor(
                self.config, self.contextMonitor, predictorName)
        elif predictorClass == 'MemorizePredictor':
            predictor = MemorizePredictor(
                self.config, self.contextMonitor, predictorName)
        elif predictorClass == 'DictionaryPredictor':
            predictor = DictionaryPredictor(
                self.config, self.contextMonitor, predictorName)
        else:
            predictor = None
        if predictor:
            self.append(predictor)

    def close_databases(self):
        """Close every opened predictors database."""
        for predictor in self:
            predictor.close_database()


class Predictor(object):
    """Base class for predictors.

    G{classtree Predictor}
    """

    __metaclass__ = ABCMeta

    def __init__(self, config, contextMonitor):
        """Predictor creator.

        @param config:
            The config is used to retrieve the predictor settings from the
            config file.
        @type config: L{Configuration}
        @param contextMonitor:
            The contextMonitor is needed because it allow the predictor to get
            the input buffers tokens.
        @type contextMonitor: L{ContextMonitor}
        """
        self.contextMonitor = contextMonitor
        self.name = "Predictor dosen't set any name"
        self.config = config

    @abstractmethod
    def predict(self, maxPartialPredictionSize, stopList):
        raise NotImplementedError("Method must be implemented")

    @abstractmethod
    def learn(self, text):
        raise NotImplementedError("Method must be implemented")


class WeightNgramPredictor(Predictor):    # observer.Observer
    """Compute prediction from n-gram model in database.

    G{classtree WeightNgramPredictor}
    """

    def __init__(self, config, contextMonitor, predictorName=None):
        """WeightNgramPredictor creator.

        @param config:
            The config is used to retrieve the predictor settings from the
            config file.
        @type config: L{drvr.Configuration}
        @param contextMonitor:
            The contextMonitor is needed because it allow the predictor to get
            the input buffers tokens.
        @type contextMonitor: L{ContextMonitor}
        @param predictorName:
            The custom name of the configuration using this predictor.
        @type predictorName: str
        """
        Predictor.__init__(self, config, contextMonitor)
        self.name = predictorName
        self.db = None
        self.dbFile = self.config.getas(self.name, 'database')
        self.deltas = self.config.getas(self.name, 'DELTAS', 'floatlist')
        self.learnMode = self.config.getas(self.name, 'learn')
        self.maxN = len(self.deltas)
        self.init_database_connector()

    def init_database_connector(self):
        """Initialize the database connector.

        Using the database file path, the n-gram maximum size and the learn
        mode to initialize and open the database.
        """
        if self.dbFile and self.maxN > 0:
            self.db = SqliteDatabaseConnector(self.dbFile, self.maxN)

    def predict(self, maxPartialPredictionSize, stopList=[]):
        """Predict the next word according to the current context.

        Use the input buffers (thanks to contextMonitor) and the n-gram database
        to predict the most probable suggestions.
        A suggestion is a word which can:
            - Predict the end of the world. i.e. complete the actual partial
              word (the user has not finished to input the word, we try to
              predict the end of the word).
            - Predict the next word (the user has type a separator after a word,
              we try to predict the next word before he starts to type it).

        In order to compute the suggestions, this method:
            - Retrieve the last n tokens from the left input buffer ; where n is
              the maximum n-grams size (max(n)) which is stored in the database.
            - Loop for each n-gram size from max(n) to 1:
                - Find n-grams of current n-gram size in the database which
                  match the last input tokens.
                - Add each retrieved n-gram to the suggestion list if it is not
                  already in it and if we have not reach the maximum number of
                  suggestions yet.

        @param maxPartialPredictionSize:
            Maximum number of suggestion to compute. If this number is reached,
            the suggestions list is immediatly return.
            DatabaseConnector.ngram_table_tp() returns the records in descending
            order according to their number of occurences so the most probable
            suggestions will be added to the list first.
            This result in no suggestion quality loss, regardless of the desired
            number of suggestions.
        @type maxPartialPredictionSize: int
        @param stopList:
            The stoplist is a list of undesirable words. Any suggestion which
            is in the stopList won't be added to the suggestions list.
        @type stopList: list

        @return:
            A list of every suggestions possible (limited to
            maxPartialPredictionSize).
        @rtype: L{Prediction}
        """
        tokens = [''] * self.maxN
        for i in range(self.maxN):
            tokens[self.maxN - 1 - i] = self.contextMonitor.left_token(i)
        prefixCompletionCandidates = []
        for k in reversed(range(self.maxN)):
            if len(prefixCompletionCandidates) >= maxPartialPredictionSize:
                break
            prefixNgram = tokens[(len(tokens) - k - 1):]
            partial = None
            partial = self.db.ngram_table_tp(
                prefixNgram,
                maxPartialPredictionSize - len(prefixCompletionCandidates))
            for p in partial:
                if len(prefixCompletionCandidates) > maxPartialPredictionSize:
                    break
                candidate = p[-2]
                if candidate not in prefixCompletionCandidates:
                    if not candidate.lower() in stopList:
                        prefixCompletionCandidates.append(candidate)
        return self.weight(prefixCompletionCandidates, tokens)

    def weight(self, prefixCompletionCandidates, tokens):
        """Compute probability of suggestions and return the most probable ones.

        The probability of a suggestion is based on its relative frequency
        toward the whole set of suggestions and the number of single tokens in
        the database.

        @param prefixCompletionCandidates:
            List of every suggestions returned by self.predict().
        @type prefixCompletionCandidates: list
        @param tokens:
            The last input tokens.
        @type tokens: list

        @return:
            List of every "good enought" suggestions.
        @rtype: L{Prediction}
        """
        prediction = Prediction()
        unigramCountsSum = self.db.sum_ngrams_occ(1)
        for j, candidate in enumerate(prefixCompletionCandidates):
            tokens[self.maxN - 1] = candidate
            probability = 0
            for k in range(self.maxN):
                numerator = self.count(tokens, 0, k + 1)
                denominator = unigramCountsSum
                if numerator > 0:
                    denominator = self.count(tokens, -1, k)
                frequency = 0
                if denominator > 0:
                    frequency = float(numerator) / denominator
                probability += self.deltas[k] * frequency
            if probability > 0:
                prediction.add_suggestion(
                    Suggestion(tokens[self.maxN - 1], probability))
        return(prediction)

    def close_database(self):
        """Close the predictor's database."""
        self.close_database()

    def learn(self, change):
        """Learn what need to be learnt by adding n-grams in database.

        @param change:
            The part of the left input buffer which represent the last change.
        @type change: str
        """
        if self.learnMode is False:
            return
        ngramMap = self.make_ngram_map(change)
        ngramMap = self.prefix_ngrams_with_input(change, ngramMap)
        self.push_ngrams_in_db(ngramMap)

    def make_ngram_map(self, change):
        """Create a map associating n-grams (lists of words) and their count.

        @param change:
            The part of the left input buffer which represent the last change.
        @type change: str
        """
        ngramMap = {}
        for curCard in range(1, self.maxN + 1):
            changeIdx = 0
            changeSize = len(change)
            ngramList = ()
            for i in range(curCard - 1):
                if changeIdx >= changeSize:
                    break
                ngramList = ngramList + (change[changeIdx],)
                changeIdx += 1
            while changeIdx < changeSize:
                ngramList = ngramList + (change[changeIdx],)
                changeIdx += 1
                try:
                    ngramMap[ngramList] = ngramMap[ngramList] + 1
                except KeyError:
                    ngramMap[ngramList] = 1
                ngramList = ngramList[1:]
            curCard += 1
        return ngramMap

    def prefix_ngrams_with_input(self, change, ngramMap):
        """Use the input left buffer to expand the n-gram map.

        This method call L{cntxt.ContextMonitor.previous_tokens} to get the
        tokens from the left input buffer that are just before the change
        and add them BEFORE the change n-grams generated by
        L{self.make_ngram_map}.

        For instance, if the current left input buffer is::
            "phone is on the white table "

        And change is::
            "table"

        Then, the n-gram map generated by self.make_ngram_map() will be::
            {("table"): 1}

        The n-gram map contain a sinle n-gram of size 1. And so this method
        will add the tokens preceding the change in the left input buffer to
        form n-grams of size 2 and more (until it reaches self.maxN)::
            {("the", "white", "table"): 1, ("white", "table"): 1, {"table"): 1}

        @param change:
            The part of the left input buffer which represent the last change.
        @type change: str
        @param ngramMap:
            Dictionary associating n-grams with their number of occurences,
            generated by self.make_ngram_map().
        @type ngramMap: dict

        @return:
            The extanded n-grams dictionary.
        @rtype: dict
        """
        changeMatchInput = (change and
                            change[-1] == self.contextMonitor.left_token(1) and
                            self.contextMonitor.left_token(len(change)))
        if changeMatchInput:
            ngramList = tuple(change[:1])
            tkIdx = 1
            while len(ngramList) < self.maxN:
                extraToken = self.contextMonitor.previous_tokens(
                    tkIdx, change)
                if not extraToken:
                    break
                ngramList = (extraToken,) + ngramList
                try:
                    ngramMap[ngramList] = ngramMap[ngramList] + 1
                except KeyError:
                    ngramMap[ngramList] = 1
                tkIdx += 1
        return ngramMap

    def push_ngrams_in_db(self, ngramMap):
        """Update the database with the n-grams contained in the n-gram map.

        Each n-gram of the n-gram map is pushed into the database with its
        number of occurences (count).
        If the n-gram is already in the database then its count (number of
        occurences) is updated. If the n-gram is not in the database then it is
        simply inserted in it.

        @param ngramMap:
            Dictionary associating n-grams with their number of occurences,
            generated by L{self.make_ngram_map} and modified by
            L{self.prefix_ngrams_with_input}.
        @type ngramMap: dict
        """
        for ngram in ngramMap:
            count = self.db.ngram_count(ngram)
            if count > 0:
                self.db.update_ngram(ngram, count + ngramMap[ngram])
            else:
                self.db.insert_ngram(list(ngram), ngramMap[ngram])
        self.db.commit()

    def count(self, tokens, offset, n):
        """Make an n-gram then retrieve and return its 'count' entry in the 

        @param tokens:
            The tokens used to make the n-gram.
        @type tokens: list
        @param offset:
            Offsset of the first token in the tokens.
        @type offset: int
        @param n:
            Size of the n-gram.
        @type n: int
        """
        if n > 0:
            ngram = tokens[len(tokens) - n + offset:len(tokens) + offset]
            result = self.db.ngram_count(ngram)
        else:
            result = self.db.sum_ngrams_occ(1)
        return result


class LastOccurPredictor(Predictor):
    """Compute predictions based on their last occurences and frequencies.

    G{classtree LastOccurPredictor}
    """

    def __init__(self, config, contextMonitor, predictorName=None):
        """LastOccurPredictor creator.

        @param config:
            The config is used to retrieve the predictor settings from the
            config file.
        @type config: L{drvr.Configuration}
        @param contextMonitor:
            The contextMonitor is needed because it allow the predictor to get
            the input buffers tokens.
        @type contextMonitor: L{ContextMonitor}
        @param predictorName:
            The custom name of the configuration using this predictor.
        @type predictorName: str
        """
        Predictor.__init__(self, config, contextMonitor)
        self.name = predictorName
        self.lambdav = self.config.getas(self.name, 'lambda', 'int')
        self.n0 = self.config.getas(self.name, 'n_0', 'int')
        self.cutoffThreshold = self.config.getas(
            self.name, 'cutoff_threshold', 'int')

    def predict(self, maxPartialPredictionSize, stopList=[]):
        """Compute the predictions using a simple exponential decay method.

        @param maxPartialPredictionSize:
            Maximum number of suggestion to compute. If this number is reached,
            the suggestions list is immediatly return.
            DatabaseConnector.ngram_table_tp() returns the records in descending
            order according to their number of occurences so the most probable
            suggestions will be added to the list first.
            This result in no suggestion quality loss, regardless of the desired
            number of suggestions.
        @type maxPartialPredictionSize: int
        @param stopList:
            The stoplist is a list of undesirable words. Any suggestion which
            is in the stopList won't be added to the suggestions list.
        @type stopList: list

        @return:
            A list of every suggestions possible (limited to
            maxPartialPredictionSize).
        @rtype: L{Prediction}
        """
        result = Prediction()
        prefix = self.contextMonitor.prefix()
        if prefix:
            index = 1
            token = self.contextMonitor.left_token(index)
            prob = 0
            while (token and
                   len(result) < maxPartialPredictionSize and
                   index <= self.cutoffThreshold):
                if token.startswith(prefix):
                    if not token.lower() in stopList:
                        prob = self.n0 * exp(- (self.lambdav * (index - 1)))
                        result.add_suggestion(Suggestion(token, prob))
                index += 1
                token = self.contextMonitor.left_token(index)
        return result

    def learn(self, text):
        """This predictor has no ability to learn."""
        pass


class MemorizePredictor(Predictor):
    """Predict words based on memorized (learnt) input tokens patterns.

    This predictor is capable of tokens memorization. It memorize the inputed
    tokens and try to predict the suggestion using memorized tokens and n-grams
    (group of consecutive tokens).

    G{classtree MemorizePredictor}
    """

    def __init__(self, config, contextMonitor, predictorName=None):
        """MemorizePredictor creator.

        @param config:
            The config is used to retrieve the predictor settings from the
            config file.
        @type config: L{drvr.Configuration}
        @param contextMonitor:
            The contextMonitor is needed because it allow the predictor to get
            the input buffers tokens.
        @type contextMonitor: L{ContextMonitor}
        @param predictorName:
            The custom name of the configuration using this predictor.
        @type predictorName: str
        """
        Predictor.__init__(self, config, contextMonitor)
        self.name = predictorName
        self.memory = self.config.getas(self.name, 'memory')
        self.trigger = self.config.getas(self.name, 'trigger', 'int')
        self.learnMode = self.config.getas(self.name, 'learn')

    def predict(self, maxPartialPredictionSize, stopList):
        """Predict words based on memorized input tokens.

        @param maxPartialPredictionSize:
            Maximum number of suggestion to compute. If this number is reached,
            the suggestions list is immediatly return.
            DatabaseConnector.ngram_table_tp() returns the records in descending
            order according to their number of occurences so the most probable
            suggestions will be added to the list first.
            This result in no suggestion quality loss, regardless of the desired
            number of suggestions.
        @type maxPartialPredictionSize: int
        @param stopList:
            The stoplist is a list of undesirable words. Any suggestion which
            is in the stopList won't be added to the suggestions list.
        @type stopList: list

        @return:
            A list of every suggestions possible (limited to
            maxPartialPredictionSize).
        @rtype: L{Prediction}
        """
        result = Prediction()
        memTrigger = []
        try:
            memFile = open(self.memory, 'r+')
        except FileNotFoundError:
            lg_error('Cannot open file ' + self.memory)
            return
        if self.init_mem_trigg(memTrigger):
            rollingWindow = ''
            if self.init_rolling_window(rollingWindow, memFile):
                token = ''
                while memFile.write(token):
                    if memTrigger == rollingWindow:
                        if not token.lower() in stopList:
                            result.add_suggestion(Suggestion(token, 1.))
                    self.update_rolling_window(rollingWindow, token)
        memFile.close()
        return result

    def learn(self, change):
        """Learn what need to be learnt by tokens in the memory file.

        @param change:
            The part of the left input buffer which represent the last change.
        @type change: str
        """
        if self.learnMode is False:
            return
        try:
            memFile = open(self.memory, 'a')
        except FileNotFoundError:
            lg_error('Cannot open file ' + self.memory)
            return
        for tok in change:
            memFile.write(tok + '\n')
        memFile.close()

    def init_mem_trigg(self, memTrigger):
        result = False
        for i in range(self.trigger, 0, -1):
            memTrigger.append(self.contextMonitor.left_token(i))
        if not '' in memTrigger:
            result = True
        return result

    def init_rolling_window(self, rollingWindow, memFile):
        tmp = [x.strip('\n') for x in memFile.readlines()]
        token = ''
        count = 0
        while count < self.trigger and tmp[count]:
            count += 1
        return count == self.trigger

    def update_rolling_window(self, rollingWindow, token):
        rollingWindow = rollingWindow[1:]
        rollingWindow += token


class DictionaryPredictor(Predictor):
    """Very simple word predictor using a dictionary.

    The dictionary is a file containing one word per line. This predictor does
    not use n-grams and is therefore less effective than the predictors using
    n-grams because it does not consider context.

    G{classtree DictionaryPredictor}
    """

    def __init__(self, config, contextMonitor, predictorName):
        """DictionaryPredictor creator.

        @note: The string.lower() and string.strip() methods have a great impact
            on performance (the profile module show that they require almost
            1 second of processing time when calculating suggestions for 10
            contexts. So this constructor no more directly use the dictionary
            file. A database is created instead.
            Every words of the dictionary are lowered and stripped then added
            to the database.
            Doing so, the performance of the predictor are way better.
            Profiling a script querying suggestions for 10 successive contexts
            show the improvement profits:
                - lower()ing and strip()ping each word of the file on each
                  predict() call::
                      ncalls  tottime  percall  cumtime  percall filename:lineno
                      690048    0.468    0.000    0.468    0.000 :0(lower)
                - Creating an improved list upon initialization and using it on
                  each predict() call (previous optimization method)::
                      ncalls  tottime  percall  cumtime  percall filename:lineno
                      100046    0.059    0.000    0.059    0.000 :0(lower)
                  It is approx. 800% faster. But this profiling mix
                  initialization and later computation. It means than most of
                  the time of the previous profiling line is spend in
                  initializing the list, computation on each predict() call are
                  even more profitable.
                - Creating a database and querying it on each predict() call::
                      ncalls  tottime  percall  cumtime  percall filename:lineno
                      100046    0.059    0.000    0.059    0.000 :0(lower)
                  It is not faster than the previous method but the database
                  must only be created once. And once it is created the
                  initialization time is (near) null and the querying time on
                  each predict() call is even faster.

        @change:
            - 08/06/15: Method now create an ordered optimized list containing
                dictionary words upon initialization in order to increase the
                speed of the predictor.
            - 13/06/15: Method now use a database containing the dictionary
                words. See: L{minr.DictMiner}

        @param config:
            The config is used to retrieve the predictor settings from the
            config file.
        @type config: L{drvr.Configuration}
        @param contextMonitor:
            The contextMonitor is needed because it allow the predictor to get
            the input buffers tokens.
        @type contextMonitor: L{ContextMonitor}
        @param predictorName:
            The custom name of the configuration using this predictor.
        @type predictorName: str
        """
        Predictor.__init__(self, config, contextMonitor)
        self.name = predictorName
        self.dbFile = self.config.getas(self.name, 'database')
        self.db = None
        self.prob = self.config.getas(self.name, 'probability', 'float')
        self.init_database_connector()

    def init_database_connector(self):
        """Initialize the database connector.

        Using the database file path, the n-gram maximum size and the learn
        mode to initialize and open the database.
        """
        if self.dbFile:
            self.db = SqliteDatabaseConnector(self.dbFile)

    def get_dict_range(self, prefix):
        """Select the dictionary range where words starts with the given prefix.

        A suggested word must complete the given token, it means that suggested
        words all start with this token, here called the prefix.
        This method create a list containing the suggested words for the
        given prefix, i.e. every words of the dictionary list starting with
        the prefix.
        It is easy as the dictionary list is ordered. For instance:

        If the prefix is::
            'hell'

        And the dictionary list is::
            ['bird', 'blue', 'given', 'hair', 'hellish', 'hello', 'red', 'zip']

        We first remove every words of the list one by one until we reach a word
        which actualy starts with the prefix 'hell', then we have::
            ['hellish', 'hello', 'red', 'zip']

        Finaly we scan every words of the remaining list and when we reach a
        word which does not starts with the given prefix then we know that every
        remaining words won't start with the prefix neither as the list is
        ordered, so we have::
            ['hellish', 'hello']

        @warning: This method has become useless since the words are now stored
                  in a database.

        @param prefix:
            The prefix from which suggested words range is computed.
        @type prefix: str
        """
        rangeWords = []
        for word in self.dictWords:
            if word.startswith(prefix):
                rangeWords = self.dictWords[self.dictWords.index(word):]
                break
        for word in rangeWords:
            if not word.startswith(prefix):
                rangeWords = rangeWords[:rangeWords.index(word)]
                break
        return rangeWords

    def predict(self, maxPartialPredictionSize, stopList):
        """Complete the actual word or predict the next word using dictionary.

        Use the input buffers (thanks to contextMonitor) and the word dictionary
        to predict the most probable suggestions.
        A suggestion is a word which can:
            - Predict the end of the world. i.e. complete the actual partial
              word (the user has not finished to input the word, we try to
              predict the end of the word).
            - Predict the next word (the user has type a separator after a
              word, we try to predict the next word before he starts to type
              it).

        In order to compute the suggestions, this method:
            - Retrieve the last token from the left input buffer.
            - Loop for each word in the dictionary:
                - If the word starts with the last token retrieved: add it to
                  the suggestion list if we have not reach the maximum number of
                  suggestions yet.
                  It is not necessary to check if the word is already in the
                  suggestion list because in a dictionary a word should only
                  appear once. In any case, the merger will merge the
                  duplicate suggestions.

        @param maxPartialPredictionSize:
            Maximum number of suggestion to compute. If this number is reached,
            the suggestions list is immediatly return.
            DatabaseConnector.ngram_table_tp() returns the records in descending
            order according to their number of occurences so the most probable
            suggestions will be added to the list first.
            This result in no suggestion quality loss, regardless of the desired
            number of suggestions.
        @type maxPartialPredictionSize: int
        @param stopList:
            The stoplist is a list of undesirable words. Any suggestion which
            is in the stopList won't be added to the suggestions list.
        @type stopList: list

        @return:
            A list of every suggestions possible (limited to
            maxPartialPredictionSize).
        @rtype: L{Prediction}
        """
        result = Prediction()
        prefix = self.contextMonitor.prefix().lower()
        count = 0
        candidates = self.db.ngram_table_tp([prefix], maxPartialPredictionSize)
        for candidate in candidates:
            if count > maxPartialPredictionSize:
                break
            candidate = candidate[-2]
            if not candidate in stopList:
                result.add_suggestion(Suggestion(candidate, self.prob))
                count += 1
        return result

    def learn(self, text):
        """This predictor has no ability to learn."""
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""The classes used for data mining.

Curently there is two data miners:
    - Corpus miner: mine from text files.
    - Facebook miner: mine from a facebook profile. In fact, this miner write
      the profile's posts into a file and then the file is mines the same way as
      a text corpus. That's why the two miners classes subclass the TextMiner
      class.

@todo 0.1.0:
    Implement the twitter miner which should be very similar to the facebook
    miner.
"""

from tipy.tknz import TextTokenizer
from tipy.db import SqliteDatabaseConnector, insert_ngrams
from os import system
from abc import ABCMeta, abstractmethod
try:
    from urllib.request import Request, urlopen
except ImportError:
    from urllib2 import Request, urlopen
from json import loads, dumps
from re import compile
from requests import request, HTTPError, get
from time import mktime
from datetime import datetime
from tipy.lg import lg


FB_CREATION_TIME = 1075852860
"""@var: Facebook creation date in unix time."""


# This class is taken from the facebook python sdk package.
# The facebook python sdk can't be build for python 3 so I modify it to make it
# compatible.
# See: https://facebook-sdk.readthedocs.org/
class GraphAPI(object):
    """A client for the Facebook Graph API.
    See http://developers.facebook.com/docs/api for complete
    documentation for the API.
    The Graph API is made up of the objects in Facebook (e.g., people,
    pages, events, photos) and the connections between them (e.g.,
    friends, photo tags, and event RSVPs). This client provides access
    to those primitive types in a generic way. For example, given an
    OAuth access token, this will fetch the profile of the active user
    and the list of the user's friends:
       graph = facebook.GraphAPI(access_token)
       user = graph.get_object("me")
       friends = graph.get_connections(user["id"], "friends")
    You can see a list of all of the objects and connections supported
    by the API at http://developers.facebook.com/docs/reference/api/.
    You can obtain an access token via OAuth or by using the Facebook
    JavaScript SDK. See
    http://developers.facebook.com/docs/authentication/ for details.
    If you are using the JavaScript SDK, you can use the
    get_user_from_cookie() method below to get the OAuth access token
    for the active user from the cookie saved by the SDK.
    """

    def __init__(self, access_token=None, timeout=None, version=None):
        # The default version is only used if the version kwarg does not exist.
        default_version = "2.0"
        valid_API_versions = ["2.0", "2.1", "2.2", "2.3"]

        self.access_token = access_token
        self.timeout = timeout

        if version:
            version_regex = compile("^\d\.\d$")
            match = version_regex.search(str(version))
            if match is not None:
                if str(version) not in valid_API_versions:
                    raise GraphAPIError("Valid API versions are " +
                                        str(valid_API_versions).strip('[]'))
                else:
                    self.version = "v" + str(version)
            else:
                raise GraphAPIError("Version number should be in the"
                                    " following format: #.# (e.g. 2.0).")
        else:
            self.version = "v" + default_version

    def get_object(self, id, **args):
        """Fetchs the given object from the graph."""
        return self.request(self.version + "/" + id, args)

    def get_objects(self, ids, **args):
        """Fetchs all of the given object from the graph.
        We return a map from ID to object. If any of the IDs are
        invalid, we raise an exception.
        """
        args["ids"] = ",".join(ids)
        return self.request(self.version + "/", args)

    def get_connections(self, id, connection_name, **args):
        """Fetchs the connections for given object."""
        return self.request(
            self.version + "/" + id + "/" + connection_name, args)

    def put_object(self, parent_object, connection_name, **data):
        """Writes the given object to the graph, connected to the given parent.
        For example,
            graph.put_object("me", "feed", message="Hello, world")
        writes "Hello, world" to the active user's wall. Likewise, this
        will comment on a the first post of the active user's feed:
            feed = graph.get_connections("me", "feed")
            post = feed["data"][0]
            graph.put_object(post["id"], "comments", message="First!")
        See http://developers.facebook.com/docs/api#publishing for all
        of the supported writeable objects.
        Certain write operations require extended permissions. For
        example, publishing to a user's feed requires the
        "publish_actions" permission. See
        http://developers.facebook.com/docs/publishing/ for details
        about publishing permissions.
        """
        assert self.access_token, "Write operations require an access token"
        return self.request(
            self.version + "/" + parent_object + "/" + connection_name,
            post_args=data,
            method="POST")

    def put_wall_post(self, message, attachment={}, profile_id="me"):
        """Writes a wall post to the given profile's wall.
        We default to writing to the authenticated user's wall if no
        profile_id is specified.
        attachment adds a structured attachment to the status message
        being posted to the Wall. It should be a dictionary of the form:
            {"name": "Link name"
             "link": "http://www.example.com/",
             "caption": "{*actor*} posted a new review",
             "description": "This is a longer description of the attachment",
             "picture": "http://www.example.com/thumbnail.jpg"}
        """
        return self.put_object(profile_id, "feed", message=message,
                               **attachment)

    def put_comment(self, object_id, message):
        """Writes the given comment on the given post."""
        return self.put_object(object_id, "comments", message=message)

    def put_like(self, object_id):
        """Likes the given post."""
        return self.put_object(object_id, "likes")

    def delete_object(self, id):
        """Deletes the object with the given ID from the graph."""
        self.request(self.version + "/" + id, method="DELETE")

    def delete_request(self, user_id, request_id):
        """Deletes the Request with the given ID for the given user."""
        self.request("%s_%s" % (request_id, user_id), method="DELETE")

    def put_photo(self, image, album_path="me/photos", **kwargs):
        """
        Upload an image using multipart/form-data.
        image - A file object representing the image to be uploaded.
        album_path - A path representing where the image should be uploaded.
        """
        return self.request(
            self.version + "/" + album_path,
            post_args=kwargs,
            files={"source": image},
            method="POST")

    def get_version(self):
        """Fetches the current version number of the Graph API being used."""
        args = {"access_token": self.access_token}
        try:
            response = request("GET",
                                        "https://graph.facebook.com/" +
                                        self.version + "/me",
                                        params=args,
                                        timeout=self.timeout)
        except HTTPError as e:
            response = loads(e.read())
            raise GraphAPIError(response)

        try:
            headers = response.headers
            version = headers["facebook-api-version"].replace("v", "")
            return float(version)
        except Exception:
            raise GraphAPIError("API version number not available")

    def request(
            self, path, args=None, post_args=None, files=None, method=None):
        """Fetches the given path in the Graph API.
        We translate args to a valid query string. If post_args is
        given, we send a POST request to the given path with the given
        arguments.
        """
        args = args or {}

        if post_args is not None:
            method = "POST"

        if self.access_token:
            if post_args is not None:
                post_args["access_token"] = self.access_token
            else:
                args["access_token"] = self.access_token

        try:
            response = request(method or "GET",
                                        "https://graph.facebook.com/" +
                                        path,
                                        timeout=self.timeout,
                                        params=args,
                                        data=post_args,
                                        files=files)
        except HTTPError as e:
            response = loads(e.read())
            raise GraphAPIError(response)

        headers = response.headers
        if 'json' in headers['content-type']:
            result = response.json()
        elif 'image/' in headers['content-type']:
            mimetype = headers['content-type']
            result = {"data": response.content,
                      "mime-type": mimetype,
                      "url": response.url}
        elif "access_token" in parse_qs(response.text):
            query_str = parse_qs(response.text)
            if "access_token" in query_str:
                result = {"access_token": query_str["access_token"][0]}
                if "expires" in query_str:
                    result["expires"] = query_str["expires"][0]
            else:
                raise GraphAPIError(response.json())
        else:
            raise GraphAPIError('Maintype was not text, image, or querystring')

        if result and isinstance(result, dict) and result.get("error"):
            raise GraphAPIError(result)
        return result

    def fql(self, query):
        """FQL query.
        Example query: "SELECT affiliations FROM user WHERE uid = me()"
        """
        return self.request(self.version + "/" + "fql", {"q": query})

    def get_app_access_token(self, app_id, app_secret):
        """Get the application's access token as a string."""
        args = {'grant_type': 'client_credentials',
                'client_id': app_id,
                'client_secret': app_secret}

        return self.request("oauth/access_token", args=args)["access_token"]

    def get_access_token_from_code(
            self, code, redirect_uri, app_id, app_secret):
        """Get an access token from the "code" returned from an OAuth dialog.
        Returns a dict containing the user-specific access token and its
        expiration date (if applicable).
        """
        args = {
            "code": code,
            "redirect_uri": redirect_uri,
            "client_id": app_id,
            "client_secret": app_secret}

        return self.request("oauth/access_token", args)

    def extend_access_token(self, app_id, app_secret):
        """
        Extends the expiration time of a valid OAuth access token. See
        <https://developers.facebook.com/roadmap/offline-access-removal/
        #extend_token>
        """
        args = {
            "client_id": app_id,
            "client_secret": app_secret,
            "grant_type": "fb_exchange_token",
            "fb_exchange_token": self.access_token,
        }

        return self.request("oauth/access_token", args=args)


class MinerRegistry(list):
    """List every "miner" classes to be used.

    MinerRegistry gather every miners to be used according to the configuration
    file. It provide methods to mine corpuses from different sources, using
    different processing strategies and store the results in different outputs.

    G{classtree MinerRegistry}
    """

    def __init__(self, config):
        """ Constructor of the MinerRegistry class.

        @param config:
            The configuration file. It is used to retrieve the miners classes
            names that will be added to the MinerRegistry.
        @type config: L{drvr.Configuration}
        """
        self.config = config
        self.set_miners()   # TODO: remove (never evaluated)

    def set_miners(self):
        """Add miner class(es) to the list according to the configuration.

        Retrieve the miners classes names from the configuration and try to
        add them to the list.
        """
        self[:] = []
        for miner in self.config.getas('MinerRegistry', 'miners', 'list'):
            self.add_miner(miner)

    def add_miner(self, minerName):
        """Try to add a miner class to the list using its name.

        Get the miner class name from the configuration and create an instance
        of this class if it exists, then, add the instance to the list.

        @param minerName:
            The name of the miner. It must correspond to a section of the
            configuration so that its miner class can be retrieved.
        @type minerName: str
        """
        try:
            minerClass = self.config[minerName]['class']
        except KeyError:
            print("Config file is incorrect. Check the miner key of the"
                  " %s section if exists or create it." % (minerName))
        if minerClass == 'CorpusMiner':
            miner = CorpusMiner(self.config, minerName)
        elif minerClass == 'FacebookMiner':
            miner = CorpusMiner(self.config, minerName)
        else:
            print('WARNING: miner class "%s" is unknown. miner won\'t be'
                  'added to the registry.' % minerName)
            miner = None
        if miner:
            self.append(miner)

    def mine(self):
        """Use the miner instances list to mine the sources.

        Loop through every miner instances of the list and call their mine()
        method to perform their mining operation.
        """
        for miner in self:
            miner.mine()

    def close_databases(self):
        """Close database of every miner instances using a database."""
        for miner in self:
            if callable(getattr(miner, "close_database", None)):
                miner.close_database()


class Miner(object):
    """Abstract class for all miners.

    G{classtree Miner}
    """

    __metaclass__ = ABCMeta

    def __init__(self, config=None, minerName='None', callback=None):
        self.config = config
        self.name = minerName
        self.callback = callback
        self.dbFile = self.config.getas(self.name, 'database')

    def rm_db(self):
        """Remove the database file (call os.system)."""
        system("rm %s" % (self.dbFile))

    @abstractmethod
    def mine(self):
        raise NotImplementedError("Method must be implemented")


class TextMiner(Miner):
    """The miner for text files.

    This miner mines text files by extracting valid n-grams from them and
    inserting them in databases. Mining a text require:
        - Tokenizing the text.
        - Extracting n-grams.
        - Inserting n-grams in a database in a special way.

    @see: L{TextTokenizer}, L{db.insert_ngrams}

    G{classtree TextMiner}
    """

    def __init__(self, config, minerName, callback=None):
        """Constructor of the FacebookMiner class.

        @param config:
            The configuration file. It is used to retrieve the miner parameters.
        @type config: L{drvr.Configuration}
        @param minerName:
            The name of the miner.
        @type minerName: str
        @param callback:
            The callback is used to show the progress percentage. In the gui a
            callback method is implemented to update a progress bar showing the
            n-grams insertion progress (cf. py).
        @type callback: fun(float, ...)
        """
        super().__init__(config, minerName, callback)
        self.lowercase = self.config.getas(self.name, 'lowercase')
        self.n = self.config.getas(self.name, 'n', 'int')

    def update_db(self, textPath):
        """Mine a text file, updating the database.

        @param textPath:
            The path to the text file to mine.
        @type textPath: str
        """
        for i in range(1, self.n + 1):
            self.add_to_db(self.crt_ngram_map(textPath, i), i, True)

    def crt_new_db(self, textPath):
        """Mine a text file.

        This method dosen't try to update the n-grams counts so it will fail if
        it tries to add an n-gram which is already in the database but this
        method is a little faster than update_db().

        @note: If you're intending to create a new database but it already
               exists please consider calling rm_db() first.

        @param textPath:
            The path to the text file to mine.
        @type textPath: str
        """
        for i in range(1, self.n + 1):
            self.add_to_db(self.crt_ngram_map(textPath, i), i, False)

    def crt_ngram_map(self, textPath, n):
        """Create a n-gram dictionary from a file.

        @param textPath:
            The path to the text file to mine.
        @type textPath: str
        @param n:
            The n in n-gram. Specify the maximum size of the n-grams to
            generate.
        @type n: int

        @return:
            The n-gram dictionary.
        @rtype: dict
        """
        lg.info("Parsing " + str(n) + "-grams from " + textPath)
        self.callback(0, 'parsing ' + str(n) + '-grams from ' + textPath)
        self.tokenizer = TextTokenizer(
            textPath, n, self.lowercase, 0, self.callback)
        ngramMap = self.tokenizer.tknize_text()
        lg.info(
            str(len(ngramMap)) + ' ngrams have been extracted from ' + textPath)
        return ngramMap

    def add_to_db(self, ngramMap, n, append=False):
        """Add n-grams of an n-gram dictionary to the database.

        @param ngramMap:
            The n-gram dictionnary returned by TextTokenizer.tknize_text().
            See the above-mentioned method docstring for more information.
        @type ngramMap: dict
        @param n:
            The n in n-gram. Specify the maximum size of the n-grams to
            generate.
        @type n: int
        @param append:
            Indicate weither the n-grams should be appened to the database.
        @type append: bool
        """
        lg.info("Writing result to " + self.dbFile)
        self.callback(0, 'writing ' + str(n) + '-grams to ' + self.dbFile)
        insert_ngrams(ngramMap, n, self.dbFile, append, True, self.callback)
        lg.info('n-grams successfully added to the database')


class CorpusMiner(TextMiner):
    """The miner for text corpus.

    This miner is basically a L{minr.TextMiner} wrapper that implement the
    mine() method which merely loops on every files of the corpus and call the
    L{minr.TextMiner.update_db} method to effectively do the mining operation.

    G{classtree CorpusMiner}
    """

    def __init__(self, config, minerName, callback=None):
        """Constructor of the CorpusMiner class.

        @param config:
            The configuration file. It is used to retrieve the miner parameters.
        @type config: L{drvr.Configuration}
        @param minerName:
            The name of the miner.
        @type minerName: str
        @param callback:
            The callback is used to show the progress percentage. In the gui a
            callback method is implemented to update a progress bar showing the
            n-grams insertion progress (cf. py).
        @type callback: fun(float, ...)
        """
        super().__init__(config, minerName, callback)
        self.callback = callback
        self.corpusFiles = self.config.getas(self.name, 'texts', 'list')

    def mine(self):
        """Perform the mining operation."""
        for text in self.corpusFiles:
            self.update_db(text)
        self.callback(100, 'Done')


class DictMiner(Miner):
    """A miner to mine dictionary-like files.

    This miner isn't a real miner as it only extract words from a
    dictionary-like file and insert them into a database.
    A dictionnary-like file is a file listing words, one word per line::
        about
        army
        bath
        boat
        ...

    G{classtree DictMiner}
    """

    def __init__(self, config, minerName, callback=None):
        """Constructor of the DictMiner class.

        @param config:
            The configuration file. It is used to retrieve the miner parameters.
        @type config: L{drvr.Configuration}
        @param minerName:
            The name of the miner.
        @type minerName: str
        @param callback:
            The callback is used to show the progress percentage. In the gui a
            callback method is implemented to update a progress bar showing the
            n-grams insertion progress (cf. py).
        @type callback: fun(float, ...)
        """
        super().__init__(config, minerName, callback)
        self.dictFile = self.config.getas(self.name, 'dictionary')
        self.database = self.config.getas(self.name, 'database')

    def mine(self):
        """Perform the mining operation.

        @note: This method could have used the update_db() method like the C
            orpusMiner and FbMiner do but this method avoid useless operations
            and is, therefore, faster.

        @todo 0.0.9:
            Make sure every lines of the file contain one single word (or none).
        """
        progress = 0
        sql = SqliteDatabaseConnector(self.database)
        sql.crt_ngram_table()
        with open(self.dictFile) as dictFile:
            for i, l in enumerate(dictFile):
                pass
        noLines = i + 1
        with open(self.dictFile) as dictFile:
            for word in dictFile:
                ngram = [word.strip('\n').lower()]
                oldCount = sql.ngram_count(ngram)
                if oldCount > 0:
                    sql.update_ngram(ngram, oldCount + 1)
                else:
                    sql.insert_ngram(ngram, 1)
                progress += 100 / noLines
                if self.callback:
                    self.callback(progress)
        sql.commit()
        sql.crt_index(1)
        sql.close_database()

    def insert_words(self):
        progress = 0
        sql = SqliteDatabaseConnector(self.database)
        sql.crt_ngram_table()
        with open(self.dictFile) as dictFile:
            for i, l in enumerate(dictFile):
                pass
        noLines = i + 1
        with open(self.dictFile) as dictFile:
            for word in dictFile:
                ngram = [word.strip('\n').lower()]
                oldCount = sql.ngram_count(ngram)
                if oldCount > 0:
                    sql.update_ngram(ngram, oldCount + 1)
                else:
                    sql.insert_ngram(ngram, 1)
                progress += 100 / noLines
                if self.callback:
                    self.callback(progress)
        sql.commit()
        sql.crt_index(1)
        sql.close_database()


class FacebookMiner(TextMiner):
    """The Facebook user profile miner.

    This miner use an access token to access a user facebook wall and retrieve
    its text message. While the messages's text is retrieve, the miner write
    them into a txt file. Once every messages have been write to the file the
    miner generate n-grams from the file and insert them in the database, using
    the L{minr.TextMiner} methods.

    G{classtree FacebookMiner}

    @note: The miner do not retrieve every facebook wall messages each time.
        When mining a facebook wall he saves the published date of the latest
        message and on next mining it will only retrieved the messages that have
        been published AFTER the saved date.
        See: L{minr.FacebookMiner.update_fb}

    @todo 0.2.0:
        Create web app in order to log the user to facebook and twitter,
        authenticate them and ask for permissions (if needed) and finaly get
        access token.
    """

    def __init__(self, config, minerName, callback=None):
        """Constructor of the FacebookMiner class.

        @param config:
            The configuration file. It is used to retrieve the miner parameters.
        @type config: L{drvr.Configuration}
        @param minerName:
            The name of the miner.
        @type minerName: str
        @param callback:
            The callback is used to show the progress percentage. In the gui a
            callback method is implemented to update a progress bar showing the
            n-grams insertion progress (cf. py).
        @type callback: fun(float, ...)
        """
        super().__init__(config, minerName, callback)
        self.fbFile = 'fb.txt'
        self.callback = callback
        self.accessToken = self.config.getas(self.name, 'accesstoken')
        self.previousLast = self.config.getas(self.name, 'last_update', 'int')

    def mine(self):
        """Perform the mining operation."""
        if not self.accessToken is None:
            try:
                jsonText = self.get_user_details()
                jsonText = loads(jsonText)
                userInfo = loads(jsonText)
            except TypeError:
                self.callback(0, 'error: Faebook access token is invalid')
                lg.error('Facebook access token is invalid')
                return
        else:
            self.callback(0, 'error: Faebook access token is missing')
            lg.error('Facebook access token is missing')
            return
        self.callback(0, 'Mining facebook user posts')
        PAGE_ID = userInfo['id']
        graph = GraphAPI(self.accessToken)
        profile = graph.get_object(PAGE_ID)
        posts = graph.get_connections(profile['id'], 'posts')
        self.update_fb(posts)
        self.update_db(self.fbFile)
        self.callback(100, 'Done')

    def write_to_file(self, post, fo):
        """Write a facebook post message to a file.

        @param post:
            A facebook post is a dictionary. If the post contains a textual
            message then it is associated to the 'message' key.
        @type post: dict
        @param fo:
            The file to write in.
        @type fo: TextIOWrapper
        """
        if "message" in post:
            fo.write(post['message'].encode('utf-8') + "\n".encode('ascii'))

    def scrap_fb(self, posts):
        """Fully mine every posts of a facebook profile.

        Mine the posts contained in the "posts" parameter (which should contains
        the latest posts) and keep requesting older posts until we reach the
        last post. If a post contains a textual message it is automatically
        written in the file at self.fbFile.
        The latest post unix time is computed and written in the config.

        @note: It is not possible to get every posts of a facebook profile in a
            single request (except if the profile contains very few posts).So
            The method must scan the "posts" dictionary and request the older
            posts until the request return a posts dictionary.

        @param posts:
            "posts" is returned by GraphAPI.get_connections(), it contains
            posts of a facebook profile.
        @type posts: dict
        """
        fo = open(self.fbFile, "wb")
        lastPostTime = ''
        while True:
            try:
                if lastPostTime == '':
                    lastPostTime = mktime(datetime.strptime(
                        posts['data'][0]['created_time'],
                        "%Y-%m-%dT%H:%M:%S+0000").timetuple())
                for post in posts['data']:
                    self.write_to_file(post, fo)
                posts = get(posts['paging']['next']).json()
            except KeyError:
                break
        fo.close()
        if lastPostTime:
            self.config[self.name]['LAST_UPDATE'] = \
                str(int(float(lastPostTime)))

    def update_fb(self, posts):
        """Mine posts of a facebook profile since the last mining operation.

        Mine the posts contained in the "posts" parameter (which should contains
        the latest posts) and keep requesting older posts until:
            - We reach a post that has already be mined (comparison is carried
              out using unix time and the 'last_update' config option)
            - We reach the last post of the facebook profile.
        If a post contains a textual message it is automatically written in
        the file at self.fbFile.
        The latest post unix time is computed and written in the config so that
        we know which posts have been published after this one the next time
        the method is called.

        @note: It is not possible to get every posts of a facebook profile in a
               single request (except if the profile contains very few posts).
               So The method must scan the "posts" dictionary and request the
               older posts until the request return a posts dictionary.

        @param posts:
            "posts" is returned by GraphAPI.get_connections(), it contains posts
            of a facebook profile.
        @type posts: dict
        """
        fo = open(self.fbFile, "wb")
        lastPostTime = ''
        stop = False
        if self.previousLast:
            previousLast = self.previousLast
        else:
            previousLast = FB_CREATION_TIME
        while True:
            try:
                if lastPostTime == '':
                    lastPostTime = mktime(datetime.strptime(
                        posts['data'][0]['created_time'],
                        "%Y-%m-%dT%H:%M:%S+0000").timetuple())
                    diff = lastPostTime - previousLast
                else:
                    try:
                        # not very regular, particularly on first mining
                        self.callback(float(100 * previousLast / mktime(
                            datetime.strptime(
                                posts['data'][0]['created_time'],
                                "%Y-%m-%dT%H:%M:%S+0000").timetuple())))
                    except IndexError:
                        pass
                for post in posts['data']:
                    postTime = mktime(datetime.strptime(
                        post['created_time'],
                        "%Y-%m-%dT%H:%M:%S+0000").timetuple())
                    if postTime > previousLast:
                        self.write_to_file(post, fo)
                    else:
                        stop = True
                        break
                if stop:
                    break
                posts = get(posts['paging']['next']).json()
            except KeyError:
                break
        fo.close()
        if lastPostTime:
            self.config[self.name]['last_update'] = \
                str(int(float(lastPostTime)))

    def get_user_details(self):
        """Use the facebook access token to get details about the user.

        @return:
            The user details or an empty dictionary if the request fail wich
            probably means that the access token is invalid or outdated.
        @rtype: dict
        """
        jDict = {}
        url = "https://graph.facebook.com/me?access_token=" + self.accessToken
        try:
            response = urlopen(Request(url))
            jDict = dumps(response.read().decode('utf-8'))
        except Exception:
            pass
        return jDict

    def rm_db(self):
        """Override the parent method.

        This method delete the database file and also set the last_update option
        of the facebook miner to the oldest value possible so that the facebook
        account will be fully scraped on next mining operation.
        """
        system("rm %s" % (self.dbFile))
        self.config['FbMiner']['last_update'] = str(FB_CREATION_TIME)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__version__ = '0.0.8'
__date__ = '06/2015'
__author__ = 'Mathieu Fourcroy'
__contact__ = 'mathieu.fourcroy@gmail.com'
__copyright__ = 'Copyright (c) 2015 Mathieu Fourcroy'



from setuptools import setup
import os

setup(
    name="tipy",
    version="0.0.8",
    author="Mathieu Fourcroy",
    author_email="mathieu.fourcroy@gmail.com",
    packages=["tipy"],
    package_dir={'tipy': 'tipy'},
    classifiers=[
        'Programming Language :: Python :: 3.4',
        'Topic :: Text Processing :: Linguistic',
    ],
    keywords='text prediction competer keyboard',
    include_package_data=True,
    url="http://pypi.python.org/pypi/tipy/",
    license="GPL",
    description="Text predictor.",
    long_description=open("README.txt").read(),
    data_files=[
        (os.path.join(os.path.expanduser("~"), '.config/tipy'),
            ['cfg/tipy.ini']),
        (os.path.join(os.path.expanduser("~"), '.config/tipy/databases'), 
            ['databases/corp.db',
             'databases/dict.db',
             'databases/fb.db',
             'databases/user.db'
             ]),
        (os.path.join(os.path.expanduser("~"), '.config/tipy/dicttionaries'), 
            ['dictionaries/en_US.dic']),
        (os.path.join(os.path.expanduser("~"), '.config/tipy/stoplists'), 
            ['stoplists/insanities_en.stoplist']),
        (os.path.join(os.path.expanduser("~"), '.config/tipy/txt'), 
            ['txt/brown.txt',
             'txt/fb.txt',
             'txt/fr.txt'
            ]),
        ('/usr/share/tipy',
            ['cfg/tipy.ini']),
        ('/usr/share/tipy/databases', 
            ['databases/corp.db',
             'databases/dict.db',
             'databases/fb.db',
             'databases/user.db'
             ]),
        ('/usr/share/tipy/dicttionaries', 
            ['dictionaries/en_US.dic']),
        ('/usr/share/tipy/stoplists', 
            ['stoplists/insanities_en.stoplist']),
        ('/usr/share/tipy/txt', 
            ['txt/brown.txt',
             'txt/fb.txt',
             'txt/fr.txt'
            ]),
    ],
    entry_points = {"console_scripts": ['tipy=tipy.main:main']},
    install_requires=["requests",],
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# The following can appear on program shutdown:
# QBasicTimer::start: QBasicTimer can only be used with threads started with
# QThread
# There is nothing to worry about it is due to Python GC deleting Qt C++ objects
# in the wrong order.

# If you have any poblem with the ui such as ugly characters font or missing
# menu bar try unseting the environnement variable with :
# $ QT_QPA_PLATFORMTHEME=
# It default value is something like:
# QT_QPA_PLATFORMTHEME=appmenu-qt5
# I can't do much to solve these bugs, it's on Qt side.
# See: https://bugs.launchpad.net/ubuntu/+source/appmenu-qt5/+bug/1307619

"""The program 

@todo 0.0.9:
    Correct the input selection bug.
"""

from PyQt5 import QtCore, QtGui, QtWidgets
from logging import StreamHandler, DEBUG
from tipy.db import SqliteDatabaseConnector
from sqlite3 import OperationalError
from tipy.minr import CorpusMiner, DictMiner, FacebookMiner
from tipy.lg import (lg, CNRM, CBLK, CRED, CGRN, CYEL, CBLU, CMAG, CCYN, CWHT,
                     BNRM, BBLK, BRED, BGRN, BYEL, BBLU, BMAG, BCYN, BWHT, BOLD,
                     CF)
from distutils.version import LooseVersion


class LoggerTextEdit(QtWidgets.QTextEdit):
    """A multiline text field showing log messages."""

    def __init__(self, config):
        super(LoggerTextEdit, self).__init__()
        self.pal = QtGui.QPalette()
        textc = QtGui.QColor(255, 255, 255)
        self.pal.setColor(QtGui.QPalette.Text, textc)
        self.setPalette(self.pal)
        self.setReadOnly(True)
        font = QtGui.QFont()
        try:
            font.setPointSize(config.getas('GUI', 'font_size', 'int'))
        except KeyError:
            font.setPointSize(10)
        self.setFont(font)

    def write(self, text):
        """Simulate LogHandler by having a write() method."""
        text = text.replace(CNRM, '<font color="White">')
        text = text.replace(CBLK, '<font color="Gray">')
        text = text.replace(CRED, '<font color="Crimson">')
        text = text.replace(CGRN, '<font color="LimeGreen">')
        text = text.replace(CYEL, '<font color="Gold">')
        text = text.replace(CBLU, '<font color="RoyalBlue ">')
        text = text.replace(CMAG, '<font color="DeepPink ">')
        text = text.replace(CCYN, '<font color="MediumAquaMarine ">')
        text = text.replace(CWHT, '<font color="White">')
        text = text.replace(BNRM, '<b><font color="White"></b>')
        text = text.replace(BBLK, '<b><font color="Gray"></b>')
        text = text.replace(BRED, '<b><font color="Crimson"></b>')
        text = text.replace(BGRN, '<b><font color="LimeGreen"></b>')
        text = text.replace(BYEL, '<b><font color="Gold"></b>')
        text = text.replace(BBLU, '<b><font color="RoyalBlue "></b>')
        text = text.replace(BMAG, '<b><font color="DeepPink "></b>')
        text = text.replace(BCYN, '<b><font color="MediumAquaMarine "></b>')
        text = text.replace(BWHT, '<b><font color="White"></b>')
        text = text.replace(BOLD, '<b></b>')
        text = text.replace('\n', '<br>')
        text = text + '</font>'
        self.insertHtml(text)


class LogDockWidget(QtWidgets.QDockWidget):
    """A main window dockable widget showing the app logs."""

    def __init__(self, config):
        super(LogDockWidget, self).__init__('log')
        self.setWidget(LoggerTextEdit(config))
        self.handler = StreamHandler(self.widget())
        self.handler.setLevel(DEBUG)
        self.handler.setFormatter(CF)
        lg.addHandler(self.handler)

    def setColor(self, fg, bg):
        """Set the background color of the log window."""
        self.widget().setStyleSheet("color: %s;background-color: %s" % (fg, bg))


class MainWindow(QtWidgets.QMainWindow):
    """The main window of the 

    G{classtree MainWindow}
    """

    CORPUS_NGRAM_PREDICTOR = 0
    USER_SMOOTH_NGRAM_PREDICTOR = 1
    FB_SMOOTH_NGRAM_PREDICTOR = 2
    LAST_OCCUR_PREDICTOR = 3
    MEMORIZE_PREDICTOR = 4
    DICTIONARY_PREDICTOR = 5
    TEXT_FILE_MINER = 0
    DICTIONARY_MINER = 1
    FACEBOOK_MINER = 2
    TWITTER_MINER = 3

    def __init__(self, driver):
        """MainWindow creator.

        It allow the user to:
            - Type text and get predictive suggestions.
            - Directly complete input words or insert predicted words in the
              input.
            - See the miners informations.
            - Execute mining operations using the defined miners.
            - Delete miners database.
            - Modify almost every options of the configuration file using the
              settings window.

        @param driver:
            The Driver instance wich contains everything needed for word
            prediction.
        @type driver: L{Driver}
        """
        super(MainWindow, self).__init__()
        self.driver = driver
        self.config = self.driver.configuration
        self.predictors = [
            'CorpusNgramPredictor',
            'InputNgramPredictor',
            'FbNgramPredictor',
            'LateOccurPredictor',
            'MemorizePredictor',
            'DictionaryPredictor']
        self.miners = [
            'CorpusMiner',
            'DictMiner',
            'FbMiner',
            'TwitterMiner']
        self.prevText = ''
        self.prevCursorPosition = 0
        self.prevTextLen = 0
        self.dbPath = '[UNDEFINED]'
        self.nGramSize = '[UNDEFINED]'
        self.predsUsingDb = []
        self.setupUi(self)
        self.mainTabs.setCurrentIndex(1)
        self.on_miner_selected(self.minersComboBox.currentIndex())

    def setupUi(self, MainWindow):
        """Create and set the main window widgets and layouts."""
        MainWindow.resize(640, 510)
        MainWindow.setWindowTitle('Preditor')
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        MainWindow.setCentralWidget(self.centralwidget)
        self.mainTabs = QtWidgets.QTabWidget(self.centralwidget)
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout_7 = QtWidgets.QGridLayout()
        self.gridLayout_8 = QtWidgets.QGridLayout()
        self.gridLayout_3 = QtWidgets.QGridLayout()
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        ############################## MINING TAB ##############################
        self.miningTab = QtWidgets.QWidget()
        self.gridLayout_5 = QtWidgets.QGridLayout(self.miningTab)
        # =================== DATABASE STATISTICS GROUP BOX ================== #
        self.dbStatsGroupBox = QtWidgets.QGroupBox(self.miningTab)
        self.dbStatsGroupBox.setTitle('Miner\'s database statistics')
        self.gridLayout_6 = QtWidgets.QGridLayout(self.dbStatsGroupBox)
        # Path label
        self.dbPathLabel = QtWidgets.QLabel(self.dbStatsGroupBox)
        self.dbPathLabel.setText('Path:')
        # Path value label
        self.dbPathValueLabel = QtWidgets.QLabel(self.dbStatsGroupBox)
        self.dbPathValueLabel.setText('-')
        # n-grams size label
        self.ngramsSizeLabel = QtWidgets.QLabel(self.dbStatsGroupBox)
        self.ngramsSizeLabel.setText('N-grams size:')
        # n-grams size value label
        self.ngramsSizeValueLabel = QtWidgets.QLabel(self.dbStatsGroupBox)
        self.ngramsSizeValueLabel.setText('-')
        # Number of n-grams label
        self.noNgramsLabel = QtWidgets.QLabel(self.dbStatsGroupBox)
        self.noNgramsLabel.setText('Number of n-grams:')
        # Table associating n-grams size value and n-grams count
        self.ngramsTable = QtWidgets.QTableWidget(self.dbStatsGroupBox)
        self.ngramsTable.setColumnCount(2)
        self.ngramsTable.setHorizontalHeaderItem(
            0, QtWidgets.QTableWidgetItem())
        self.ngramsTable.setHorizontalHeaderItem(
            1, QtWidgets.QTableWidgetItem())
        self.ngramsTable.setSortingEnabled(True)
        self.ngramsTable.horizontalHeaderItem(0).setText('N-gram size')
        self.ngramsTable.horizontalHeaderItem(1).setText('Number of n-grams')
        # Predictors using the miner database label
        self.predsUsingDbLabel = QtWidgets.QLabel(self.dbStatsGroupBox)
        self.predsUsingDbLabel.setText('Predictors using the database:')
        # List of predictors using the miner database as input
        self.predsUsingDbList = QtWidgets.QListWidget(self.dbStatsGroupBox)
        # ==================== MINING OPERATIONS GROUP BOX =================== #
        self.miningGroupBox = QtWidgets.QGroupBox(self.miningTab)
        self.miningGroupBox.setTitle('Mining operations')
        self.gridLayout_2 = QtWidgets.QGridLayout(self.miningGroupBox)
        # A comboBox to select between available miners
        self.minersComboBox = QtWidgets.QComboBox(self.miningGroupBox)
        for miner in self.miners:
            self.minersComboBox.addItem(miner)
        self.minersComboBox.currentIndexChanged.connect(self.on_miner_selected)
        # Operation status label
        self.minerOperationLabel = QtWidgets.QLabel(self.miningGroupBox)
        self.minerOperationLabel.setText('Operation:')
        # Operation status value label
        self.minerOperationValueLabel = QtWidgets.QLabel(self.miningGroupBox)
        self.minerOperationValueLabel.setText('-')
        # Operation progression label
        self.progressLabel = QtWidgets.QLabel(self.miningGroupBox)
        self.progressLabel.setText('Progression:')
        # Operation progress bar
        self.progressBar = QtWidgets.QProgressBar(self.miningGroupBox)
        self.progressBar.setProperty("value", 0)
        self.progressBar.setAlignment(QtCore.Qt.AlignCenter)
        self.progressBar.setTextVisible(True)
        self.progressBar.setInvertedAppearance(False)
        self.progressBar.setTextDirection(QtWidgets.QProgressBar.TopToBottom)
        self.progressBar.setRange(0, 100)
        # Delete database file button
        self.deleteDbBtn = QtWidgets.QPushButton(self.miningGroupBox)
        self.deleteDbBtn.setText('Delete database file')
        self.deleteDbBtn.released.connect(self.on_rm_db_btn_released)
        # Run mining operation button
        self.mineBtn = QtWidgets.QPushButton(self.miningGroupBox)
        self.mineBtn.setText('MINE !')
        self.mineBtn.released.connect(self.on_mine_btn_released)
        ############################ PREDICTING TAB ############################
        self.predictingTab = QtWidgets.QWidget()
        self.gridLayout_4 = QtWidgets.QGridLayout(self.predictingTab)
        self.inputTextEdit = QtWidgets.QPlainTextEdit(self.predictingTab)
        #~ self.inputTextEdit.textChanged.connect(self.on_input_text_change)
        self.inputTextEdit.cursorPositionChanged.connect(
            self.on_input_text_cursor_position_change)
        self.suggestionsList = QtWidgets.QListWidget(self.predictingTab)
        self.suggestionsList.itemDoubleClicked.connect(
            self.on_suggestion_double_clicked)
        ######################### TABS CONTENT AND NAME ########################
        self.mainTabs.addTab(self.miningTab, '')
        self.mainTabs.addTab(self.predictingTab, '')
        self.mainTabs.setTabText(
            self.mainTabs.indexOf(self.miningTab), 'Mining')
        self.mainTabs.setTabText(
            self.mainTabs.indexOf(self.predictingTab), 'Predicting')
        ###################### ADD EVERYTHING TO LAYOUTS #######################
        self.horizontalLayout.addWidget(self.deleteDbBtn)
        self.horizontalLayout.addWidget(self.mineBtn)
        self.gridLayout_2.addWidget(self.minersComboBox, 0, 0, 1, 1)
        self.gridLayout_3.addWidget(self.minerOperationLabel, 0, 0, 1, 1)
        self.gridLayout_3.addWidget(self.progressLabel, 1, 0, 1, 1)
        self.gridLayout_3.addWidget(self.progressBar, 1, 1, 1, 1)
        self.gridLayout_3.addWidget(self.minerOperationValueLabel, 0, 1, 1, 1)
        self.gridLayout_2.addLayout(self.gridLayout_3, 4, 0, 1, 1)
        self.gridLayout_2.addLayout(self.horizontalLayout, 1, 0, 1, 1)
        self.gridLayout_3.addWidget(self.minerOperationValueLabel, 0, 1, 1, 1)
        self.gridLayout_5.addWidget(self.miningGroupBox, 0, 0, 1, 1)
        self.gridLayout_4.addWidget(self.inputTextEdit, 0, 0, 1, 1)
        self.gridLayout.addWidget(self.mainTabs, 0, 0, 1, 1)
        self.gridLayout_6.addWidget(self.noNgramsLabel, 6, 0, 1, 1)
        self.gridLayout_7.addWidget(self.dbPathLabel, 0, 0, 1, 1)
        self.gridLayout_7.addWidget(self.ngramsSizeLabel, 1, 0, 1, 1)
        self.gridLayout_7.addWidget(self.dbPathValueLabel, 0, 1, 1, 1)
        self.gridLayout_7.addWidget(self.ngramsSizeValueLabel, 1, 1, 1, 1)
        self.gridLayout_6.addLayout(self.gridLayout_7, 0, 0, 1, 1)
        self.gridLayout_6.addWidget(self.predsUsingDbLabel, 1, 0, 1, 1)
        self.gridLayout_6.addLayout(self.gridLayout_8, 2, 0, 1, 1)
        self.gridLayout_6.addWidget(self.ngramsTable, 7, 0, 1, 1)
        self.gridLayout_5.addWidget(self.dbStatsGroupBox, 3, 0, 1, 1)
        self.gridLayout_6.addWidget(self.predsUsingDbList, 2, 0, 1, 1)
        self.gridLayout_5.addWidget(self.dbStatsGroupBox, 3, 0, 1, 1)
        self.gridLayout_4.addWidget(self.suggestionsList, 1, 0, 1, 1)
        ################################ MENUBAR ###############################
        # The menubar
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 640, 18))
        # ============================= THE MENUS ============================ #
        # The File menu
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setTitle('File')
        # The Edition menu
        self.menuEdition = QtWidgets.QMenu(self.menubar)
        self.menuEdition.setTitle('Edition')
        # The View menu
        self.menuView = QtWidgets.QMenu(self.menubar)
        self.menuView.setTitle('View')
        # =========================== THE ACTIONS ============================ #
        # The Open action
        self.actionOpen = QtWidgets.QAction(MainWindow)
        self.actionOpen.setText('Import some text...')
        # The Quit action
        self.actionQuit = QtWidgets.QAction(MainWindow)
        self.actionQuit.setText('Quit')
        self.actionQuit.setShortcut('Ctrl+Q')
        self.actionQuit.triggered.connect(
            QtCore.QCoreApplication.instance().quit)
        # The Paste action
        self.actionPaste = QtWidgets.QAction(MainWindow)
        self.actionPaste.setText('Paste')
        self.actionPaste.setShortcut('Ctrl+V')
        self.actionPaste.triggered.connect(self.on_paste_btn_triggered)
        # The Copy action
        self.actionCopy = QtWidgets.QAction(MainWindow)
        self.actionCopy.setText('Copy')
        self.actionCopy.setShortcut('Ctrl+C')
        self.actionCopy.triggered.connect(self.on_copy_btn_triggered)
        # The Cut action
        self.actionCut = QtWidgets.QAction(MainWindow)
        self.actionCut.setText('Cut')
        self.actionCut.setShortcut('Ctrl+X')
        self.actionCut.triggered.connect(self.on_cut_btn_triggered)
        # The Settings action
        self.actionSettings = QtWidgets.QAction(MainWindow)
        self.actionSettings.setText('Preferences')
        self.actionSettings.setShortcut('Ctrl+Alt+P')
        self.actionSettings.triggered.connect(self.on_pref_triggered)
        # The Afficher les probabilités action
        self.actionShowProbabilities = QtWidgets.QAction(MainWindow)
        self.actionShowProbabilities.setText('Show probabilities')
        self.actionShowProbabilities.triggered.connect(
            self.on_show_probabilities_triggered)
        self.actionShowProbabilities.setCheckable(True)
        # ===================== PUTTING IT ALL TOGETHER ====================== #
        # Add the actions to the File menu
        self.menuFile.addAction(self.actionOpen)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionQuit)
        # Add the actions to the Edition menu
        self.menuEdition.addAction(self.actionCut)
        self.menuEdition.addAction(self.actionCopy)
        self.menuEdition.addAction(self.actionPaste)
        self.menuEdition.addSeparator()
        self.menuEdition.addAction(self.actionSettings)
        # Add the actions to the View menu
        self.menuView.addAction(self.actionShowProbabilities)
        # Add the menus to the menubar
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuEdition.menuAction())
        self.menubar.addAction(self.menuView.menuAction())
        # Set the menubar
        MainWindow.setMenuBar(self.menubar)
        ############################### STATUSBAR ##############################
        # The statusbar
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        # Set the statusbar
        MainWindow.setStatusBar(self.statusbar)
        ############################ LOG DOCK WIDGET ###########################
        # The dock widget for the log
        self.logDock = LogDockWidget(self.config)
        self.logDock.setColor('white', 'black')
        # Set the dock
        MainWindow.addDockWidget(QtCore.Qt.DockWidgetArea(8), self.logDock)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)
        # Show something
        lg.info('Session starts.')

    def on_paste_btn_triggered(self, checked=False):
        """Paste what's inside the clipboard."""
        self.inputTextEdit.paste()

    def on_copy_btn_triggered(self, checked=False):
        """Copy the selection."""
        self.inputTextEdit.copy()

    def on_cut_btn_triggered(self, checked=False):
        """Cut the selection."""
        self.inputTextEdit.cut()

    def on_show_probabilities_triggered(self, checked=False):
        """Show the suggested words probabilities."""
        lg.warning('Not yet implemented.')

    def on_pref_triggered(self, checked=False):
        """Open the settings dialog."""
        settingsDialog = stngs.Settings_UI(self.config)
        settingsDialog.exec_()
        self.write_config()

    def write_config(self):
        """Write the current configuration in the configuration file."""
        if self.driver.configFile:
            with open(self.driver.configFile, 'w') as f:
                self.config.write(f)

    def predictors_using_db(self):
        """Create a list of every predictors name which use a given database.

        @return:
            The list of the predictors using the database file 'self.dbPath'.
        @rtype: list
        """
        predsUsingDb = []
        for pred in self.predictors:
            try:
                if self.config[pred]['database'] == self.dbPath:
                    predsUsingDb.append(pred)
            except KeyError:
                continue
        return predsUsingDb

    def on_miner_selected(self, miner):
        """Callback called when the user select a miner in the combo box.

        When a miner is selected in the miner selection combo box some settings
        of the miner have to be retrieved from the configuration in order to
        display them to the user using the set_miner_widgets_values() method.
        As every miners store their results (n-grams) in a database the database
        path and the n-gram size are retrieved from the configuration and stored
        into the "dbPath" and "nGramSize" instance variables.

        @param miner:
            The index of the miner in the combo box.
        @type miner: int
        """
        if miner == self.TEXT_FILE_MINER:
            self.dbPath = self.config.getas('CorpusMiner', 'database')
            self.nGramSize = self.config.getas('CorpusMiner', 'n')
            self.predsUsingDb = self.predictors_using_db()
        elif miner == self.DICTIONARY_MINER:
            self.dbPath = self.config.getas('DictMiner', 'database')
            self.nGramSize = '1'
            self.predsUsingDb = self.predictors_using_db()
        elif miner == self.FACEBOOK_MINER:
            self.dbPath = self.config.getas('FbMiner', 'database')
            self.nGramSize = self.config.getas('FbMiner', 'n')
            self.predsUsingDb = self.predictors_using_db()
        elif miner == self.TWITTER_MINER:
            self.dbPath = '[UNDEFINED]'
            self.nGramSize = '[UNDEFINED]'
        else:
            self.dbPath = '[UNDEFINED]'
            self.nGramSize = '[UNDEFINED]'
        self.set_miner_widgets_values()

    def set_miner_widgets_values(self):
        """Update the widgets (labels and list) displaying infos on the miner.

        When a miner is selected in the miner selection combo box, its config
        settings are retrieve from the configuration and the instance variables
        "dbPath", "nGramSize" and "predsUsingDb" are set.
        This method modify the value of the widgets displaying the informations
        about the miner with the above-mentioned variables.
        """
        nGramCount = []
        self.dbPathValueLabel.setText(self.dbPath)
        self.ngramsSizeValueLabel.setText(self.nGramSize)
        while self.predsUsingDbList.count():
            self.predsUsingDbList.takeItem(0)
        for pred in self.predictors:
            if pred in self.predsUsingDb:
                self.predsUsingDbList.addItem(pred)
        if self.nGramSize == '[UNDEFINED]' or self.dbPath == '[UNDEFINED]':
            self.ngramsTable.setRowCount(0)
            return
        database = SqliteDatabaseConnector(self.dbPath, self.nGramSize)
        for n in range(1, int(self.nGramSize) + 1):
            try:
                nGramCount.append(database.ngrams_in_table(n))
            except OperationalError:
                nGramCount.append(0)
        database.close_database()
        self.ngramsTable.setRowCount(n)
        for i, count in enumerate(nGramCount):
            self.ngramsTable.setItem(
                i, 0, QtWidgets.QTableWidgetItem(str(i + 1)))
            self.ngramsTable.setItem(
                i, 1, QtWidgets.QTableWidgetItem(str(count)))

    def progress_callback(self, perc=0, text=None):
        """Update the label showing the mining operation and the progress bar.

        This method is a callback which is called from the miners. It update the
        text of the label displaying the minig operation and the value of the
        progress bar displaying the operation progression.

        @param perc:
            Computed by the miners calling the callback, it indicate the
            operation progression and is usually a float but it must be
            converted to int as the progress bar only show integers.
        @type perc: float or int
        @param text:
            The text to display in the mining operation label.
        @type text: str
        """
        if text:
            self.minerOperationValueLabel.setText(text)
            self.minerOperationValueLabel.repaint()
        self.progressBar.setValue(round(perc))

    def on_rm_db_btn_released(self):
        """Callback called when the user press the "Delete database" button.

        The method first identify the current miner selected in the combo box
        using its index then create an instance of this miner and called its
        rm_db() method to effectively carry out the database suppression
        operation.
        Some miners modify the configuration so the config file needs to be
        rewrite after the operation.
        Here is a short description of the miners database suppression
        operation, please refer to their rm_db() method docstring for more
        informations:
            - CorpusMiner: Remove the database file.
            - FbMiner: Remove the database file and set the "last_update" config
              option to the minimum value so the config have to be writen
              afterward.
        """
        if self.minersComboBox.currentIndex() == self.TEXT_FILE_MINER:
            miner = CorpusMiner(self.config, 'CorpusMiner')
            miner.rm_db()
        elif self.minersComboBox.currentIndex() == self.DICTIONARY_MINER:
            miner = DictMiner(self.config, 'DictMiner')
            miner.rm_db()
        elif self.minersComboBox.currentIndex() == self.FACEBOOK_MINER:
            miner = FacebookMiner(self.config, 'FbMiner')
            miner.rm_db()
            self.write_config()
        elif self.minersComboBox.currentIndex() == self.TWITTER_MINER:
            self.progress_callback(0, 'error: [MINER NOT IMPLEMENTED YET]')
            lg.error('Miner not implemented yet')
            return
        else:   # should never happen
            self.progress_callback(0, 'error: unknown miner')
            lg.error('Unknown miner "{0}"'.format(
                self.minersComboBox.currentIndex()))
            return
        self.set_miner_widgets_values()

    def on_mine_btn_released(self):
        """Callback called when the user press the "MINE!" button.

        The method first identify the current miner selected in the combo box
        using its index then create an instance of this miner and called its
        mine() method to effectively carry out the mining operation.
        Some miners modify the configuration so the config file needs to be
        rewrite after the operation.
        Here is a short description of the miners mining operation, please refer
        to their mine() method docstring for more informations:
            - CorpusMiner: Mine a text corpus (i.e. a set of text files) by
              extracting n-grams from the files and inserting them into a
              database.
            - FbMiner: Mine a facebook user wall (only text messages from
              posts) by extracting n-grams from the posts and inserting them
              into a database. This miner modifies the "last_update" option of
              its config section so the config have to be writen afterward.
            - TwiterMiner: Not implemented yet but it will be very similar to
              FbMiner.
        """
        if self.minersComboBox.currentIndex() == self.TEXT_FILE_MINER:
            miner = CorpusMiner(
                self.config, 'CorpusMiner', self.progress_callback)
            miner.mine()
        elif self.minersComboBox.currentIndex() == self.DICTIONARY_MINER:
            miner = DictMiner(self.config, 'DictMiner', self.progress_callback)
            miner.mine()
        elif self.minersComboBox.currentIndex() == self.FACEBOOK_MINER:
            miner = FacebookMiner(
                self.config, 'FbMiner', self.progress_callback)
            miner.mine()
            self.write_config()
        elif self.minersComboBox.currentIndex() == self.TWITTER_MINER:
            self.progress_callback(0, 'error: [MINER NOT IMPLEMENTED YET]')
            lg.warning('Miner not implemented yet')
            return
        else:   # should never happen
            self.progress_callback(0, 'error: unknown miner')
            lg.warning('Unknown miner')
            return
        self.set_miner_widgets_values()

    def on_input_text_change(self):
        """Update the input buffers and compute the suggestion.

        This method is called whenever one or more character(s) is/are added to
        or removed from the input text. It update the input left and right
        buffers according to the text change and then generate the suggestions
        for the new input context.
        There is three king of input text change:
            - Characters have been appened:
              -> Characters are added to the left input buffer (via
              callback.update()).
            - Part of the characters have been removed:
              -> Characters are removed to the left input buffer (via
              callback.update()). It simulates a backspace input.
            - Every characters have been removed (while new ones have been
              added):
              -> Every characters of the left input buffer are removed (via
              callback.update()). New characters, if any, are added to the left
              input buffer (via callback.update()).
        @note: It correspond to a Ctrl+A then <some printable characters> or
            <backspace>.
        """
        text = self.inputTextEdit.toPlainText()
        if text.startswith(self.prevText):
            change = text[len(self.prevText):]
            self.driver.callback.update(change)
        elif self.prevText.startswith(text):
            self.driver.callback.update('\b', len(self.prevText) - len(text))
        else:
            self.driver.callback.update('\b', len(self.prevText))
            self.driver.callback.update(text)
        self.prevText = text
        self.make_suggestions()

    def make_suggestions(self):
        """Compute and show suggestions.

        Request the PredictorActivator to compute the suggestions and show them
        in the suggestion list.
        """
        self.suggestionsList.clear()
        for p in self.driver.predict():
            self.suggestionsList.addItem(p)

    def on_input_text_cursor_position_change(self):
        """Callback called when the input text cursor position change.

        There's three possible moves types:
            - The cursor moved because one or more character(s) have been added
              to or removed from the input text:
              -> The cursor position and the input text length are saved and the
              suggestions are updated according to the current input.
            - The cursor moved because the user pressed the left arrow key or
              clicked somewhere in the text, on the left of the previous cursor
              position:
              -> In this case the input text has not change we call the Driver
              callback to modify the input left and right buffer so that the
              left buffer now contains what's on the left of the cursor and the
              right buffer now contains what's on the right of the cursor. The
              suggestions are then updated according to the current cursor
              position.
            - The cursor moved because the user pressed the right arrow key or
              clicked somewhere in the text, on the right of the previous cursor
              position:
              -> In this case the input text has not change we call the Driver
              callback to modify the input left and right buffer so that the
              left buffer now contains what's on the left of the cursor and the
              right buffer now contains what's on the right of the cursor. The
              suggestions are then updated according to the current cursor
              position.

        @note: The on_input_text_change() method used to be connected to the
            textChanged signal of the input text widget but it is redundant
            because if the text change, the cursor position automatically change
            too. It save some operations.

        @bug:
            When selecting characters in the input field (self.inputTextEdit)
            the cursor position is modified and induce erroneous context changes
            and word predictions.
        """
        currentCursorPosition = self.inputTextEdit.textCursor().position()
        currentTextLen = len(self.inputTextEdit.toPlainText())
        posDiff = currentCursorPosition - self.prevCursorPosition
        lenDiff = currentTextLen - self.prevTextLen
        if lenDiff == posDiff:              # Char(s) have been added or removed
            self.on_input_text_change()
        elif posDiff < 0 and lenDiff == 0:  # Move cursor to the left
            self.driver.callback.update('\x1B[D', posDiff)
        elif posDiff > 0 and lenDiff == 0:  # Move cursor to the right
            self.driver.callback.update('\x1B[C', posDiff)
        else:                               # Should never happen
            self.prevCursorPosition = currentCursorPosition
            self.prevTextLen = currentTextLen
            return
        self.prevCursorPosition = currentCursorPosition
        self.prevTextLen = currentTextLen
        self.make_suggestions()

    def on_suggestion_double_clicked(self, item):
        """Complete the input with the selected suggestion.

        When the user double click on a word of the list the word is used to
        complete the input text. There is two kind of completion depending on
        the input cursor position:
            - The cursor is at the end of the text:
              -> The completion for the last token is computed and append to the
              input text. A space is then added allowing the user to save a key
              and the program to immediatly suggest the next word prediction.
            - The cursor is not at the end of the text:
              -> The completion for the token in which the cursor is is computed
              and the end of the token is replaced by the completion.

        During this method the input text widget signals are disconnected which
        means that no suggestions will be compute during the completion process.
        This is important because this method will modify the input text but, of
        course, there is no need to generate any suggestions as the user isn't
        typing anything during the process.

        @note: A token can be a word or a separator. If it is a word then the
               completion represents the end of the word. If it is a separator
               then the completion represents the next word.

        @param item:
            The selected item (word) in the suggestion list.
        @type item:
            QtWidgets.QListWidgetItem
        """
        self.inputTextEdit.cursorPositionChanged.disconnect()
        currentCursorPosition = self.inputTextEdit.textCursor().position()
        if currentCursorPosition == len(self.inputTextEdit.toPlainText()):
            completion = self.driver.make_completion(item.text())
            # If the word is finished then suggestion = word and completion = ''
            # But we still have to add a space after the word so check if
            # completion is not False
            if not completion is False:
                self.inputTextEdit.insertPlainText(completion)
                self.driver.callback.update(completion)
                self.prevCursorPosition = \
                    self.inputTextEdit.textCursor().position()
                self.prevTextLen = len(self.inputTextEdit.toPlainText())
                self.prevText = self.inputTextEdit.toPlainText()
                self.inputTextEdit.cursorPositionChanged.connect(
                    self.on_input_text_cursor_position_change)
                self.inputTextEdit.insertPlainText(' ')
        else:
            completion = self.driver.make_completion(item.text())
            if completion:
                suffix = self.driver.contextMonitor.suffix()
                tmpCursor = self.inputTextEdit.textCursor()
                tmpCursor.movePosition(
                    QtQTextCursor.Right,
                    QtQTextCursor.MoveAnchor, len(suffix))
                self.inputTextEdit.setTextCursor(tmpCursor)
                self.driver.callback.update('\x1B[C', len(suffix))
                for i in range(len(suffix)):
                    self.inputTextEdit.textCursor().deletePreviousChar()
                    self.driver.callback.update('\b', 1)
                self.inputTextEdit.insertPlainText(completion)
                self.driver.callback.update(completion)
                self.prevCursorPosition = \
                    self.inputTextEdit.textCursor().position()
                self.prevTextLen = len(self.inputTextEdit.toPlainText())
                self.prevText = self.inputTextEdit.toPlainText()
        self.inputTextEdit.cursorPositionChanged.connect(
            self.on_input_text_cursor_position_change)
